# ******************************************************************************
# Copyright (c) 2019 Analog Devices, Inc.  All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# - Redistributions of source code must retain the above copyright notice, this
#  list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice,
#  this list of conditions and the following disclaimer in the documentation
#  and/or other materials provided with the distribution.
# - Modified versions of the software must be conspicuously marked as such.
# - This software is licensed solely and exclusively for use with
#  processors/products manufactured by or for Analog Devices, Inc.
# - This software may not be combined or merged with other code in any manner
#  that would cause the software to become subject to terms and conditions
#  which differ from those listed here.
# - Neither the name of Analog Devices, Inc. nor the names of its contributors
#  may be used to endorse or promote products derived from this software
#  without specific prior written permission.
# - The use of this software may or may not infringe the patent rights of one
#  or more patent holders.  This license does not release you from the
#  requirement that you obtain separate licenses from these patent holders to
#  use this software.
#
# THIS SOFTWARE IS PROVIDED BY ANALOG DEVICES, INC. AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# NONINFRINGEMENT, TITLE, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL ANALOG DEVICES, INC. OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, PUNITIVE OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, DAMAGES ARISING OUT OF
# CLAIMS OF INTELLECTUAL PROPERTY RIGHTS INFRINGEMENT; PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# ******************************************************************************

from typing import List, Dict

from ..core import utils
from ..core.enums.dcb_enums import DCBCommand
from .common_application import CommonApplication
from ..core.enums.ad7156_enums import AD7156Command
from ..core.packets.command_packet import CommandPacket
from ..core.enums.common_enums import Application, CommonCommand
from ..core.packets.common_packets import DCBPacket, RegisterPacket


class AD7156Application(CommonApplication):
    """
    AD7156 Application class.

    .. code-block:: python3
        :emphasize-lines: 4

        from adi_study_watch import SDK

        sdk = SDK("COM4")
        application = sdk.get_ad7156_application()
    """

    def __init__(self, packet_manager):
        super().__init__(Application.AD7156, packet_manager)
        self._dcb_size = 20

    def delete_device_configuration_block(self) -> Dict:
        """
        Deletes AD7156 Device configuration block.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_ad7156_application()
            application.delete_device_configuration_block()
        """
        packet = DCBPacket(self._destination, DCBCommand.ERASE_CONFIG_REQ)
        packet.set_dcb_data_size(0)
        return self._send_packet(packet, DCBCommand.ERASE_CONFIG_RES)

    def load_configuration(self) -> Dict:
        """
        Loads configuration.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_ad7156_application()
            x = application.load_configuration()
            print(x["payload"]["status"])
            # CommonStatus.OK
        """
        packet = CommandPacket(self._destination, AD7156Command.LOAD_CONFIG_REQ)
        return self._send_packet(packet, AD7156Command.LOAD_CONFIG_RES)

    def read_device_configuration_block(self) -> Dict:
        """
        Returns entire device configuration block.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_ad7156_application()
            x = application.read_device_configuration_block()
            print(x["payload"]["dcb_data"])
            # []
        """
        packet = DCBPacket(self._destination, DCBCommand.READ_CONFIG_REQ)
        packet.set_dcb_data_size(0)
        return self._send_packet(packet, DCBCommand.READ_CONFIG_RES)

    def read_register(self, addresses: List[int]) -> Dict:
        """
        Reads the register values of specified addresses. This function takes a list of addresses to read,
        and returns a response packet as dictionary containing addresses and values.

        :param addresses: List of register addresses to read.
        :type addresses: List[int]
        :return: A response packet as dictionary
        :rtype: Dict

        .. list-table::
           :widths: 50 50
           :header-rows: 1

           * - Address Lower Limit
             - Address Upper Limit
           * - 0x00
             - 0x17

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_ad7156_application()
            x = application.read_register([0x10, 0x12])
            print(x["payload"]["data"])
            # [['0x10', '0x0'], ['0x12', '0x0']]
        """
        address_range = [0x00, 0x17]
        packet = RegisterPacket(self._destination, CommonCommand.REGISTER_READ_REQ)
        packet.set_number_of_operations(len(addresses))
        utils.check_array_address_range(addresses, address_range, 2)
        packet.set_read_reg_data(addresses)
        return self._send_packet(packet, CommonCommand.REGISTER_READ_RES)

    def set_timeout(self, timeout_value: float) -> None:
        """
        Sets the time out for queue to wait for command packet response.

        :param timeout_value: queue timeout value (in sec).
        :type timeout_value: int
        :return: None

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_ad7156_application()
            application.set_timeout(10)
        """
        super().set_timeout(timeout_value)

    def write_device_configuration_block(self, addresses_values: List[List[int]]) -> Dict:
        """
        Writes the device configuration block values of specified addresses.
        This function takes a list of addresses and values to write, and returns a response packet as
        dictionary containing addresses and values.

        :param addresses_values: List of addresses and values to write.
        :type addresses_values: List[List[int]]
        :return: A response packet as dictionary.
        :rtype: Dict

        .. list-table::
           :widths: 50 50
           :header-rows: 1

           * - Address Lower Limit
             - Address Upper Limit
           * - 0x09
             - 0x12

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_ad7156_application()
            x = application.write_device_configuration_block([[0x10, 2], [0x12, 0x1]])
            print(x["payload"]["size"])
            # 2
        """
        address_range = [0x09, 0x12]
        packet = DCBPacket(self._destination, DCBCommand.WRITE_CONFIG_REQ)
        utils.check_array_address_range(addresses_values, address_range, 2)
        packet.set_size(utils.split_int_in_bytes(len(addresses_values), length=2))
        packet.set_dcb_write_data(addresses_values, self._dcb_size, len(addresses_values))
        return self._send_packet(packet, DCBCommand.WRITE_CONFIG_RES)

    def write_device_configuration_block_from_file(self, filename: str) -> Dict:
        """
        Writes the device configuration block values of specified addresses from file.

        :param filename: dcb filename
        :return: A response packet as dictionary.
        :rtype: Dict

        .. list-table::
           :widths: 50 50
           :header-rows: 1

           * - Address Lower Limit
             - Address Upper Limit
           * - 0x09
             - 0x12

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_ad7156_application()
            application.write_device_configuration_block_from_file("adxl_dcb.dcfg")
        """
        result = self._write_device_configuration_block_from_file_helper(filename)
        if result:
            return self.write_device_configuration_block(result)

    def write_register(self, addresses_values: List[List[int]]) -> Dict:
        """
        Writes the register values of specified addresses. This function takes a list of addresses and values to write,
        and returns a response packet as dictionary containing addresses and values.

        :param addresses_values: List of register addresses and values to write.
        :type addresses_values: List[List[int]]
        :return: A response packet as dictionary.
        :rtype: Dict

        .. list-table::
           :widths: 50 50
           :header-rows: 1

           * - Address Lower Limit
             - Address Upper Limit
           * - 0x09
             - 0x12

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_ad7156_application()
            x = application.write_register([[0x10, 0x1], [0x12, 0x2]])
            print(x["payload"]["data"])
            # [['0x10', '0x1'], ['0x12', '0x2']]
        """
        address_range = [0x09, 0x12]
        packet = RegisterPacket(self._destination, CommonCommand.REGISTER_WRITE_REQ)
        packet.set_number_of_operations(len(addresses_values))
        utils.check_array_address_range(addresses_values, address_range, 2)
        packet.set_write_reg_data(addresses_values)
        return self._send_packet(packet, CommonCommand.REGISTER_WRITE_RES)



# ******************************************************************************
# Copyright (c) 2019 Analog Devices, Inc.  All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# - Redistributions of source code must retain the above copyright notice, this
#  list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice,
#  this list of conditions and the following disclaimer in the documentation
#  and/or other materials provided with the distribution.
# - Modified versions of the software must be conspicuously marked as such.
# - This software is licensed solely and exclusively for use with
#  processors/products manufactured by or for Analog Devices, Inc.
# - This software may not be combined or merged with other code in any manner
#  that would cause the software to become subject to terms and conditions
#  which differ from those listed here.
# - Neither the name of Analog Devices, Inc. nor the names of its contributors
#  may be used to endorse or promote products derived from this software
#  without specific prior written permission.
# - The use of this software may or may not infringe the patent rights of one
#  or more patent holders.  This license does not release you from the
#  requirement that you obtain separate licenses from these patent holders to
#  use this software.
#
# THIS SOFTWARE IS PROVIDED BY ANALOG DEVICES, INC. AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# NONINFRINGEMENT, TITLE, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL ANALOG DEVICES, INC. OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, PUNITIVE OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, DAMAGES ARISING OUT OF
# CLAIMS OF INTELLECTUAL PROPERTY RIGHTS INFRINGEMENT; PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# ******************************************************************************

from .command_packet import CommandPacket
from .. import utils
from ..enums.adpd_enums import ADPDDevice, ADPDSlot, Clock, ADPDLed


class ActiveSlotPacket(CommandPacket):
    """
    Packet Structure:

    .. code-block::

        {
            'header': {
                'source': <Application.ADPD: ['0xC1', '0x10']>,
                'destination': <Application.APP_USB: ['0xC7', '0x05']>,
                'length': '0xC',
                'checksum': '0x0'
            },
            'payload': {
                'command': <ADPDCommand.SET_SLOT_ACTIVE_RES: ['0x6B']>,
                'status': <CommonStatus.OK: ['0x00']>,
                'slot_num': <ADPDSlot.SLOT_A: ['0x01']>,
                'slot_enabled': False
            }
        }
    """

    def __init__(self, destination=None, command=None):
        super().__init__(destination, command)
        self._config["payload"]["slot_num"] = {"size": 1}
        self._config["payload"]["slot_enabled"] = {"size": 1, "join": True}
        self._packet["payload"]["slot_num"] = [0x00]
        self._packet["payload"]["slot_enabled"] = [0x00]

    def set_slot_num(self, slot_num):
        self._packet["payload"]["slot_num"] = slot_num.value

    def set_slot_active(self, slot_active):
        self._packet["payload"]["slot_enabled"] = slot_active

    def decode_packet(self, data):
        super().decode_packet(data)
        self._packet["payload"]["slot_num"] = ADPDSlot(self._packet["payload"]["slot_num"])
        self._packet["payload"]["slot_enabled"] = bool(self._packet["payload"]["slot_enabled"])


class AgcControlPacket(CommandPacket):
    """
    Packet Structure:

    .. code-block::

        {
            'header': {
                'source': <Application.ADPD: ['0xC1', '0x10']>,
                'destination': <Application.APP_USB: ['0xC7', '0x05']>,
                'length': '0xF',
                'checksum': '0x0'
            },
            'payload': {
                'command': <ADPDCommand.AGC_ON_OFF_RES: ['0x77']>,
                'status': <CommonStatus.OK: ['0x00']>,
                'num_ops': 2,
                'agc_data': [
                    [ <ADPDLed.LED_MWL: ['0x00']>, True ],
                    [ <ADPDLed.LED_GREEN: ['0x01']>, True ]
                ]
            }
        }
    """

    def __init__(self, destination=None, command=None):
        super().__init__(destination, command)
        self._config["payload"]["num_ops"] = {"size": 1, "join": True}
        self._packet["payload"]["num_ops"] = [0x01]
        self._packet["payload"]["agc_data"] = [0x00, 0x00]

    def set_number_of_operations(self, num_ops):
        self._packet["payload"]["num_ops"] = num_ops

    def set_fields_values(self, fields_values):
        self._packet["payload"]["agc_data"] = fields_values

    def decode_packet(self, data):
        super().decode_packet(data)
        address_value_data = data[11:]
        data = []
        for i in range((len(address_value_data) // 2)):
            start_index = i * 2
            value_data = address_value_data[start_index:start_index + 1]
            value = bool(utils.join_multi_length_packets(value_data))
            address_data = address_value_data[start_index + 1:start_index + 2]
            address = ADPDLed(address_data)
            data.append([address, value])
        self._packet["payload"]["agc_data"] = data


class AgcInfoPacket(CommandPacket):
    """
    Packet Structure:

    .. code-block::

        {
            'header': {
                'source': <Application.ADPD: ['0xC1', '0x10']>,
                'destination': <Application.APP_USB: ['0xC7', '0x05']>,
                'length': '0x61',
                'checksum': '0x0'
            },
            'payload': {
                'command': <ADPDCommand.AGC_INFO_RES: ['0x7B']>,
                'status': <CommonStatus.OK: ['0x00']>,
                'led_index': <ADPDLed.LED_GREEN: ['0x01']>,
                'ch1': [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
                'ch2': [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
                'dc0_led_current': 0,
                'tia_ch1': 0,
                'tia_ch2': 0
            }
        }
    """

    def __init__(self, destination=None, command=None):
        super().__init__(destination, command)
        self._config["payload"]["led_index"] = {"size": 1}
        self._config["payload"]["ch1"] = {"size": 40}
        self._config["payload"]["ch2"] = {"size": 40}
        self._config["payload"]["dc0_led_current"] = {"size": 2}
        self._config["payload"]["tia_ch1"] = {"size": 2}
        self._config["payload"]["tia_ch2"] = {"size": 2}

    def set_led_index(self, led_index):
        self._packet["payload"]["led_index"] = led_index.value

    def decode_packet(self, data):
        super().decode_packet(data)
        self._packet["payload"]["led_index"] = ADPDLed(self._packet["payload"]["led_index"])
        ch1 = self._packet["payload"]["ch1"]
        data = []
        for i in range(10):
            start_index = i * 4
            value = utils.join_multi_length_packets(ch1[start_index:start_index + 4])
            data.append(value)
        self._packet["payload"]["ch1"] = data
        ch2 = self._packet["payload"]["ch2"]
        data = []
        for i in range(10):
            start_index = i * 4
            value = utils.join_multi_length_packets(ch2[start_index:start_index + 4])
            data.append(value)
        self._packet["payload"]["ch2"] = data
        dc0_led_current = utils.join_multi_length_packets(self._packet["payload"]["dc0_led_current"])
        self._packet["payload"]["dc0_led_current"] = dc0_led_current
        self._packet["payload"]["tia_ch1"] = utils.join_multi_length_packets(self._packet["payload"]["tia_ch1"])
        self._packet["payload"]["tia_ch2"] = utils.join_multi_length_packets(self._packet["payload"]["tia_ch2"])


class AgcStatusPacket(CommandPacket):
    """
    Packet Structure:

    .. code-block::

        {
            'header': {
                'source': <Application.ADPD: ['0xC1', '0x10']>,
                'destination': <Application.APP_USB: ['0xC7', '0x05']>,
                'length': '0xC',
                'checksum': '0x0'
            },
            'payload': {
                'command': <ADPDCommand.AGC_STATUS_RES: ['0x7D']>,
                'status': <CommonStatus.OK: ['0x00']>,
                'agc_type': <ADPDLed.LED_GREEN: ['0x01']>,
                'agc_status': True
            }
        }
    """

    def __init__(self, destination=None, command=None):
        super().__init__(destination, command)
        self._config["payload"]["agc_type"] = {"size": 1}
        self._config["payload"]["agc_status"] = {"size": 1, "join": True}

    def set_agc_type(self, agc_type):
        self._packet["payload"]["agc_type"] = agc_type.value

    def decode_packet(self, data):
        super().decode_packet(data)
        self._packet["payload"]["agc_type"] = ADPDLed(self._packet["payload"]["agc_type"])
        if self._packet["payload"]["agc_status"] in [0, 1]:
            self._packet["payload"]["agc_status"] = bool(self._packet["payload"]["agc_status"])
        else:
            self._packet["payload"]["agc_status"] = None


class ComModePacket(CommandPacket):
    """
    Packet Structure:

    .. code-block::

        {
            'header': {
                'source': <Application.ADPD: ['0xC1', '0x10']>,
                'destination': <Application.APP_USB: ['0xC7', '0x05']>,
                'length': '0xB',
                'checksum': '0x0'
            },
            'payload': {
                'command': <ADPDCommand.COMMUNICATION_MODE_RES: ['0x69']>,
                'status': <CommonStatus.OK: ['0x00']>,
                'com_mode': 2
            }
        }
    """

    def __init__(self, destination=None, command=None):
        super().__init__(destination, command)
        self._config["payload"]["com_mode"] = {"size": 1, "join": True}
        self._packet["payload"]["com_mode"] = [0x00]

    def set_com_mode(self, com_mode):
        self._packet["payload"]["com_mode"] = com_mode


class SamplingFrequencyPacket(CommandPacket):
    """
    Packet Structure:

    .. code-block::

        {
            'header': {
                'source': <Application.ADPD: ['0xC1', '0x10']>,
                'destination': <Application.APP_USB: ['0xC7', '0x05']>,
                'length': '0xC',
                'checksum': '0x0'
            },
            'payload': {
                'command': <ADPDCommand.SET_SAMPLING_FREQUENCY_RES: ['0x73']>,
                'status': <CommonStatus.OK: ['0x00']>,
                'odr': 100
            }
        }
    """

    def __init__(self, destination=None, command=None):
        super().__init__(destination, command)
        self._config["payload"]["odr"] = {"size": 2, "join": True}
        self._packet["payload"]["odr"] = [0x00, 0x00]

    def set_odr(self, odr):
        self._packet["payload"]["odr"] = odr


class SlotPacket(CommandPacket):
    """
    Packet Structure:

    .. code-block::

        {
            'header': {
                'source': <Application.ADPD: ['0xC1', '0x10']>,
                'destination': <Application.APP_USB: ['0xC7', '0x05']>,
                'length': '0xF',
                'checksum': '0x0'
            },
            'payload': {
                'command': <ADPDCommand.GET_SLOT_RES: ['0x49']>,
                'status': <CommonStatus.OK: ['0x00']>,
                'slot_num': <ADPDSlot.SLOT_F: ['0x06']>,
                'slot_enabled': True,
                'slot_format': 3,
                'channel_num': 3
            }
        }
    """

    def __init__(self, destination=None, command=None):
        super().__init__(destination, command)
        self._config["payload"]["slot_num"] = {"size": 1}
        self._config["payload"]["slot_enabled"] = {"size": 1, "join": True}
        self._config["payload"]["slot_format"] = {"size": 2, "join": True}
        self._config["payload"]["channel_num"] = {"size": 1, "join": True}
        self._packet["payload"]["slot_num"] = [0x00]
        self._packet["payload"]["slot_enabled"] = [0x00]
        self._packet["payload"]["slot_format"] = [0x00, 0x00]
        self._packet["payload"]["channel_num"] = [0x00]

    def set_slot_num(self, slot_num):
        self._packet["payload"]["slot_num"] = slot_num.value

    def set_slot_enable(self, slot_enable):
        self._packet["payload"]["slot_enabled"] = slot_enable

    def set_slot_format(self, slot_format):
        self._packet["payload"]["slot_format"] = slot_format

    def set_channel_num(self, channel_num):
        self._packet["payload"]["channel_num"] = channel_num

    def decode_packet(self, data):
        super().decode_packet(data)
        self._packet["payload"]["slot_num"] = ADPDSlot(self._packet["payload"]["slot_num"])
        self._packet["payload"]["slot_enabled"] = bool(self._packet["payload"]["slot_enabled"])


class TestCommandPacket(CommandPacket):

    def __init__(self, destination=None, command=None):
        super().__init__(destination, command)
        self._config["payload"]["device_id"] = {"size": 2}
        self._config["payload"]["return_data"] = {"size": 12}
        self._packet["payload"]["device_id"] = [0x00, 0x00]
        self._packet["payload"]["return_data"] = [0x00] * 12

    def set_data(self, data):
        self._packet["payload"]["return_data"] = data

    def decode_packet(self, data):
        super().decode_packet(data)
        print(utils.convert_int_array_to_hex(self._packet["payload"]["device_id"]))


class ADPDConfigPacket(CommandPacket):
    """
    Packet Structure:

    .. code-block::

        {
            'header': {
                'source': <Application.ADPD: ['0xC1', '0x10']>,
                'destination': <Application.APP_USB: ['0xC7', '0x05']>,
                'length': '0x15',
                'checksum': '0x0'
            },
            'payload': {
                'command': <ADPDCommand.LOAD_CONFIG_RES: ['0x43']>,
                'status': <CommonStatus.OK: ['0x00']>,
                'device_id': <ADPDDevice.DEVICE_GREEN: ['0x28']>,
                'return_data': [ 255, 255, 31, 0, 112, 7, 0, 99, 0, 0 ]
            }
        }
    """

    def __init__(self, destination=None, command=None):
        super().__init__(destination, command)
        self._config["payload"]["device_id"] = {"size": 1}
        self._config["payload"]["return_data"] = {"size": 10}
        self._packet["payload"]["device_id"] = [0x00]

    def set_device_id(self, device_id):
        self._packet["payload"]["device_id"] = device_id.value

    def decode_packet(self, data):
        super().decode_packet(data)
        if not self._packet["payload"]["device_id"] == [0]:
            self._packet["payload"]["device_id"] = ADPDDevice(self._packet["payload"]["device_id"])


class ADPDPauseResumePacket(CommandPacket):
    """
    Packet Structure:

    .. code-block::

        {
            'header': {
                'source': <Application.ADPD: ['0xC1', '0x10']>,
                'destination': <Application.APP_USB: ['0xC7', '0x05']>,
                'length': '0x15',
                'checksum': '0x0'
            },
            'payload': {
                'command': <ADPDCommand.SET_PAUSE_RES: ['0x67']>,
                'status': <CommonStatus.ERROR: ['0x01']>,
                'device_id': <ADPDDevice.DEVICE_G_R_IR_B: ['0x2C']>,
                'pause': True
            }
        }
    """

    def __init__(self, destination=None, command=None):
        super().__init__(destination, command)
        self._config["payload"]["device_id"] = {"size": 1}
        self._config["payload"]["pause"] = {"size": 10, "join": True}
        self._packet["payload"]["device_id"] = ADPDDevice.DEVICE_G_R_IR_B.value
        self._packet["payload"]["pause"] = [0x00] * 10

    def set_device_id(self, device_id):
        self._packet["payload"]["device_id"] = device_id.value

    def set_data(self, data):
        self._packet["payload"]["pause"] = data

    def decode_packet(self, data):
        super().decode_packet(data)
        self._packet["payload"]["device_id"] = ADPDDevice(self._packet["payload"]["device_id"])
        self._packet["payload"]["pause"] = bool(self._packet["payload"]["pause"])


class ClockCalibrationPacket(CommandPacket):
    """
    Packet Structure:

    .. code-block::

        {
            'header': {
                'source': <Application.ADPD: ['0xC1', '0x10']>,
                'destination': <Application.APP_USB: ['0xC7', '0x05']>,
                'length': '0xB',
                'checksum': '0x0'
            },
            'payload': {
                'command': <ADPDCommand.CLOCK_CALIBRATION_RES: ['0x45']>,
                'status': <CommonStatus.OK: ['0x00']>,
                'clock_id': <Clock.CLOCK_1M_AND_32M: ['0x06']>
            }
        }
    """

    def __init__(self, destination=None, command=None):
        super().__init__(destination, command)
        self._config["payload"]["clock_id"] = {"size": 1}
        self._packet["payload"]["clock_id"] = [0x00]

    def set_clock_id(self, clock_id):
        self._packet["payload"]["clock_id"] = clock_id.value

    def decode_packet(self, data):
        super().decode_packet(data)
        self._packet["payload"]["clock_id"] = Clock(self._packet["payload"]["clock_id"])


class ExternalStreamData(CommandPacket):
    """
    Packet Structure:

    .. code-block::

        {
            'header':{
                'source': <Application.ADPD: ['0xC1','0x10']>,
                'destination': <Application.APP_USB: ['0xC7','0x05']>,
                'length': '0xC',
                'checksum': '0x0'
            },
           'payload':{
                'command': <ADPDCommand.EXT_ADPD_DATA_STREAM: ['0x80']>,
                'status': <CommonStatus.OK: ['0x00']>,
                'sequence_num': [ 2, 0, 0, 0 ],
                'data': [ 37, 26, 6, 0 ],
                'timestamp': [ 148, 201, 188, 2 ]
           }
        }
    """

    def __init__(self, destination=None, command=None):
        super().__init__(destination, command)
        self._config["payload"]["sequence_num"] = {"size": 4}
        self._config["payload"]["data"] = {"size": 4}
        self._config["payload"]["timestamp"] = {"size": 4}
        self._packet["payload"]["sequence_num"] = [0x00] * 4
        self._packet["payload"]["data"] = [0x00] * 4
        self._packet["payload"]["timestamp"] = [0x00] * 4

    def set_sequence_num(self, sequence_num):
        self._packet["payload"]["sequence_num"] = sequence_num

    def set_data(self, data):
        self._packet["payload"]["data"] = data

    def set_timestamp(self, timestamp):
        self._packet["payload"]["timestamp"] = timestamp


class ExternalStreamODR(CommandPacket):
    """
    Packet Structure:

    .. code-block::

        {
            'header': {
                'source': <Application.ADPD: ['0xC1', '0x10']>,
                'destination': <Application.APP_USB: ['0xC7', '0x05']>,
                'length': '0xC',
                'checksum': '0x0'
            },
            'payload': {
                'command': <ADPDCommand.SET_EXT_DATA_STREAM_ODR_RES: ['0x7F']>,
                'status': <CommonStatus.OK: ['0x00']>,
                'odr': 50
            }
        }
    """

    def __init__(self, destination=None, command=None):
        super().__init__(destination, command)
        self._config["payload"]["odr"] = {"size": 2, "join": True}
        self._packet["payload"]["odr"] = [0x00] * 2

    def set_sampling_frequency(self, sampling_frequency):
        self._packet["payload"]["odr"] = sampling_frequency



# ******************************************************************************
# Copyright (c) 2019 Analog Devices, Inc.  All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# - Redistributions of source code must retain the above copyright notice, this
#  list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice,
#  this list of conditions and the following disclaimer in the documentation
#  and/or other materials provided with the distribution.
# - Modified versions of the software must be conspicuously marked as such.
# - This software is licensed solely and exclusively for use with
#  processors/products manufactured by or for Analog Devices, Inc.
# - This software may not be combined or merged with other code in any manner
#  that would cause the software to become subject to terms and conditions
#  which differ from those listed here.
# - Neither the name of Analog Devices, Inc. nor the names of its contributors
#  may be used to endorse or promote products derived from this software
#  without specific prior written permission.
# - The use of this software may or may not infringe the patent rights of one
#  or more patent holders.  This license does not release you from the
#  requirement that you obtain separate licenses from these patent holders to
#  use this software.
#
# THIS SOFTWARE IS PROVIDED BY ANALOG DEVICES, INC. AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# NONINFRINGEMENT, TITLE, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL ANALOG DEVICES, INC. OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, PUNITIVE OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, DAMAGES ARISING OUT OF
# CLAIMS OF INTELLECTUAL PROPERTY RIGHTS INFRINGEMENT; PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# ******************************************************************************

import logging
from typing import List, Dict, Callable, Tuple

from ..core import utils
from .csv_logging import CSVLogger
from .common_stream import CommonStream
from ..core.enums.dcb_enums import DCBCommand
from ..core.packets.command_packet import CommandPacket
from ..core.packets.stream_data_packets import EDADataPacket
from ..core.enums.common_enums import Application, Stream, CommonCommand
from ..core.enums.eda_enums import EDACommand, EDADFTWindow, ScaleResistor
from ..core.packets.common_packets import DecimationFactorPacket, DFTPacket
from ..core.packets.eda_packets import DynamicScalingPacket, ResistorTIACalibratePacket
from ..core.packets.common_packets import DCBPacket, VersionPacket, LibraryConfigReadWritePacket

logger = logging.getLogger(__name__)


class EDAApplication(CommonStream):
    """
    EDA Application class.

    .. code-block:: python3
        :emphasize-lines: 4

        from adi_study_watch import SDK

        sdk = SDK("COM4")
        application = sdk.get_eda_application()

    """
    DFT_WINDOW_4 = EDADFTWindow.DFT_WINDOW_4
    DFT_WINDOW_8 = EDADFTWindow.DFT_WINDOW_8
    DFT_WINDOW_16 = EDADFTWindow.DFT_WINDOW_16
    DFT_WINDOW_32 = EDADFTWindow.DFT_WINDOW_32

    SCALE_RESISTOR_110 = ScaleResistor.SCALE_RESISTOR_110
    SCALE_RESISTOR_1K = ScaleResistor.SCALE_RESISTOR_1K
    SCALE_RESISTOR_2K = ScaleResistor.SCALE_RESISTOR_2K
    SCALE_RESISTOR_3K = ScaleResistor.SCALE_RESISTOR_3K
    SCALE_RESISTOR_4K = ScaleResistor.SCALE_RESISTOR_4K
    SCALE_RESISTOR_6K = ScaleResistor.SCALE_RESISTOR_6K
    SCALE_RESISTOR_8K = ScaleResistor.SCALE_RESISTOR_8K
    SCALE_RESISTOR_10K = ScaleResistor.SCALE_RESISTOR_10K
    SCALE_RESISTOR_12K = ScaleResistor.SCALE_RESISTOR_12K
    SCALE_RESISTOR_16K = ScaleResistor.SCALE_RESISTOR_16K
    SCALE_RESISTOR_20K = ScaleResistor.SCALE_RESISTOR_20K
    SCALE_RESISTOR_24K = ScaleResistor.SCALE_RESISTOR_24K
    SCALE_RESISTOR_30K = ScaleResistor.SCALE_RESISTOR_30K
    SCALE_RESISTOR_32K = ScaleResistor.SCALE_RESISTOR_32K
    SCALE_RESISTOR_40K = ScaleResistor.SCALE_RESISTOR_40K
    SCALE_RESISTOR_48K = ScaleResistor.SCALE_RESISTOR_48K
    SCALE_RESISTOR_64K = ScaleResistor.SCALE_RESISTOR_64K
    SCALE_RESISTOR_85K = ScaleResistor.SCALE_RESISTOR_85K
    SCALE_RESISTOR_96K = ScaleResistor.SCALE_RESISTOR_96K
    SCALE_RESISTOR_100K = ScaleResistor.SCALE_RESISTOR_100K
    SCALE_RESISTOR_120K = ScaleResistor.SCALE_RESISTOR_120K
    SCALE_RESISTOR_128K = ScaleResistor.SCALE_RESISTOR_128K
    SCALE_RESISTOR_160K = ScaleResistor.SCALE_RESISTOR_160K
    SCALE_RESISTOR_196K = ScaleResistor.SCALE_RESISTOR_196K
    SCALE_RESISTOR_256K = ScaleResistor.SCALE_RESISTOR_256K
    SCALE_RESISTOR_512K = ScaleResistor.SCALE_RESISTOR_512K

    def __init__(self, callback_function, packet_manager, args):
        super().__init__(Application.EDA, Stream.EDA, packet_manager, callback_function, args)
        self._callback_function = callback_function
        self._dcb_size = 2

    def get_supported_scales(self) -> List[ScaleResistor]:
        """
        List all supported scales for EDA.

        :return: Array of scales enums.
        :rtype: List

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_eda_application()
            x = application.get_supported_scales()
            print(x)
            # [<ScaleResistor.SCALE_RESISTOR_100K: ['0x0', '0x14']>, ...]
        """
        return [self.SCALE_RESISTOR_110, self.SCALE_RESISTOR_1K, self.SCALE_RESISTOR_2K, self.SCALE_RESISTOR_3K,
                self.SCALE_RESISTOR_4K, self.SCALE_RESISTOR_6K, self.SCALE_RESISTOR_8K, self.SCALE_RESISTOR_10K,
                self.SCALE_RESISTOR_12K, self.SCALE_RESISTOR_16K, self.SCALE_RESISTOR_20K, self.SCALE_RESISTOR_24K,
                self.SCALE_RESISTOR_30K, self.SCALE_RESISTOR_32K, self.SCALE_RESISTOR_40K, self.SCALE_RESISTOR_48K,
                self.SCALE_RESISTOR_64K, self.SCALE_RESISTOR_85K, self.SCALE_RESISTOR_96K, self.SCALE_RESISTOR_100K,
                self.SCALE_RESISTOR_120K, self.SCALE_RESISTOR_128K, self.SCALE_RESISTOR_160K, self.SCALE_RESISTOR_196K,
                self.SCALE_RESISTOR_256K, self.SCALE_RESISTOR_512K]

    def _scale_helper(self, scale: ScaleResistor) -> ScaleResistor:
        """
        Confirms scale is from list of Enums.
        """
        if scale in self.get_supported_scales():
            return scale
        else:
            logger.warning(f"{scale} is not supported scale, choosing {self.get_supported_scales()[0]} "
                           f"as default scale. use get_supported_scales() to know all supported scales.")
            return self.get_supported_scales()[0]

    def get_supported_dft_windows(self) -> List[EDADFTWindow]:
        """
        List all supported dft window for EDA.

        :return: Array of dft window enums.
        :rtype: List

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_eda_application()
            x = application.get_supported_dft_windows()
            print(x)
            # [<EDADFTWindow.DFT_WINDOW_4: ['0x0']>, ... , <EDADFTWindow.DFT_WINDOW_32: ['0x3']>]
        """
        return [self.DFT_WINDOW_4, self.DFT_WINDOW_8, self.DFT_WINDOW_16, self.DFT_WINDOW_32]

    def _dft_window_helper(self, dft_window: EDADFTWindow) -> EDADFTWindow:
        """
        Confirms dft_window is from list of Enums.
        """
        if dft_window in self.get_supported_dft_windows():
            return dft_window
        else:
            logger.warning(f"{dft_window} is not supported dft window, choosing {self.get_supported_dft_windows()[0]} "
                           f"as default dft window. use get_supported_dft_windows() to know all supported dft window.")
            return self.get_supported_dft_windows()[0]

    def calibrate_resistor_tia(self, min_scale: ScaleResistor, max_scale: ScaleResistor,
                               lp_resistor_tia: ScaleResistor) -> Dict:
        """
         Calibrate Resistor Trans Impedance Amplifier.

         :param min_scale: min scale for Resistor Trans Impedance Amplifier, use get_supported_scales()
                          | to list all supported scales.
         :param max_scale: max scale for Resistor Trans Impedance Amplifier, use get_supported_scales()
                          | to list all supported scales.
         :param lp_resistor_tia: lp_resistor_tia, use get_supported_scales() to list all supported scales.
         :type min_scale: ScaleResistor
         :type max_scale: ScaleResistor
         :type lp_resistor_tia: ScaleResistor
         :return: A response packet as dictionary.
         :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_eda_application()
            x = application.get_supported_scales()
            print(x)
            # [<ScaleResistor.SCALE_RESISTOR_100K: ['0x0', '0x14']>, ... ]
            x = application.calibrate_resistor_tia(application.SCALE_RESISTOR_128K, application.SCALE_RESISTOR_256K,
                                                    application.SCALE_RESISTOR_256K)
            print(x["payload"]["status"])
            # CommonStatus.OK
         """
        min_scale = self._scale_helper(min_scale)
        max_scale = self._scale_helper(max_scale)
        lp_resistor_tia = self._scale_helper(lp_resistor_tia)
        packet = ResistorTIACalibratePacket(self._destination, EDACommand.RESISTOR_TIA_CAL_REQ)
        packet.set_min_scale(min_scale)
        packet.set_max_scale(max_scale)
        packet.set_lp_resistor_tia(lp_resistor_tia)
        packet.set_calibrated_values_count(
            [utils.join_multi_length_packets(max_scale.value) - utils.join_multi_length_packets(min_scale.value) + 1])
        return self._send_packet(packet, EDACommand.RESISTOR_TIA_CAL_RES)

    def get_decimation_factor(self) -> Dict:
        """
        Returns stream decimation factor.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_eda_application()
            x = application.get_decimation_factor()
            print(x["payload"]["decimation_factor"])
            # 1

        """
        packet = DecimationFactorPacket(self._destination, CommonCommand.GET_STREAM_DEC_FACTOR_REQ)
        packet.set_stream_address(self._stream)
        return self._send_packet(packet, CommonCommand.GET_STREAM_DEC_FACTOR_RES)

    def set_decimation_factor(self, decimation_factor: int) -> Dict:
        """
        Sets decimation factor for EDA stream.

        :param decimation_factor: decimation factor for stream
        :type decimation_factor: int
        :return: A response packet as dictionary
        :rtype: Dict

        .. list-table::
           :widths: 50 50
           :header-rows: 1

           * - Decimation Lower Limit
             - Decimation Upper Limit
           * - 0x01
             - 0x05

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_eda_application()
            x = application.set_decimation_factor(2)
            print(x["payload"]["decimation_factor"])
            # 2

        """
        decimation_factor = utils.range_and_type_check(decimation_factor, type_of=int, num_bytes=1)
        packet = DecimationFactorPacket(self._destination, CommonCommand.SET_STREAM_DEC_FACTOR_REQ)
        packet.set_stream_address(self._stream)
        packet.set_decimation_factor([decimation_factor])
        return self._send_packet(packet, CommonCommand.SET_STREAM_DEC_FACTOR_RES)

    def delete_device_configuration_block(self) -> Dict:
        """
        Deletes EDA Device configuration block.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_eda_application()
            application.delete_device_configuration_block()
        """
        packet = DCBPacket(self._destination, DCBCommand.ERASE_CONFIG_REQ)
        packet.set_dcb_data_size(0)
        return self._send_packet(packet, DCBCommand.ERASE_CONFIG_RES)

    def disable_dynamic_scaling(self) -> Dict:
        """
        Disables Dynamic scaling.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_eda_application()
            application.disable_dynamic_scaling()
        """
        packet = DynamicScalingPacket(self._destination, EDACommand.DYNAMIC_SCALE_REQ)
        packet.set_enable([0x0])
        return self._send_packet(packet, EDACommand.DYNAMIC_SCALE_RES)

    def enable_dynamic_scaling(self, min_scale: ScaleResistor, max_scale: ScaleResistor,
                               lp_resistor_tia: ScaleResistor) -> Dict:
        """
         Enables Dynamic scaling.

         :param min_scale: min scale for Resistor Trans Impedance Amplifier, use get_supported_scales()
                          | to list all supported scales.
         :param max_scale: max scale for Resistor Trans Impedance Amplifier, use get_supported_scales()
                          | to list all supported scales.
         :param lp_resistor_tia: lp_resistor_tia, use get_supported_scales() to list all supported scales.
         :type min_scale: ScaleResistor
         :type max_scale: ScaleResistor
         :type lp_resistor_tia: ScaleResistor
         :return: A response packet as dictionary.
         :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_eda_application()
            x = application.get_supported_scales()
            print(x)
            # [<ScaleResistor.SCALE_RESISTOR_100K: ['0x0', '0x14']>, ... ]
            x = application.enable_dynamic_scaling(application.SCALE_RESISTOR_128K, application.SCALE_RESISTOR_256K,
                                                    application.SCALE_RESISTOR_256K)
            print(x["payload"]["status"])
            # CommonStatus.OK

         """
        min_scale = self._scale_helper(min_scale)
        max_scale = self._scale_helper(max_scale)
        lp_resistor_tia = self._scale_helper(lp_resistor_tia)
        packet = DynamicScalingPacket(self._destination, EDACommand.DYNAMIC_SCALE_REQ)
        packet.set_min_scale(min_scale.value)
        packet.set_max_scale(max_scale.value)
        packet.set_lp_resistor_tia(lp_resistor_tia.value)
        packet.set_enable([0x1])
        return self._send_packet(packet, EDACommand.DYNAMIC_SCALE_RES)

    def get_version(self) -> Dict:
        """
        Returns EDA version info.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_ecg_application()
            x = application.get_version()
            print(x["payload"]["major_version"])
            # 3
            print(x["payload"]["minor_version"])
            # 4
            print(x["payload"]["patch_version"])
            # 3
            print(x["payload"]["version_string"])
            # EDA_App
            print(x["payload"]["build_version"])
            # TEST EDA_VERSION STRING
        """
        packet = VersionPacket(self._destination, CommonCommand.GET_VERSION_REQ)
        return self._send_packet(packet, CommonCommand.GET_VERSION_RES)

    def read_device_configuration_block(self) -> Dict:
        """
        Returns entire device configuration block.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_eda_application()
            x = application.read_device_configuration_block()
            print(x["payload"]["dcb_data"])
            # []

        """
        packet = DCBPacket(self._destination, DCBCommand.READ_CONFIG_REQ)
        packet.set_dcb_data_size(0)
        return self._send_packet(packet, DCBCommand.READ_CONFIG_RES)

    def read_library_configuration(self, fields: List[int]) -> Dict:
        """
        Reads library configuration from specified field values.

        :param fields: List of field values to read.
        :type fields: List[int]
        :return: A response packet as dictionary
        :rtype: Dict

        .. list-table::
           :widths: 50 50
           :header-rows: 1

           * - Fields Lower Limit
             - Fields Upper Limit
           * - 0x00
             - 0x02

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_eda_application()
            x = application.read_library_configuration([0x00])
            print(x["payload"]["data"])
            # [['0x0', '0x0']]

        """
        address_range = [0x00, 0x02]
        packet = LibraryConfigReadWritePacket(self._destination, CommonCommand.READ_LCFG_REQ)
        fields_size = len(fields)
        packet.set_number_of_operations(fields_size)
        utils.check_array_address_range(fields, address_range, 1)
        packet.set_read_fields_data(fields)
        return self._send_packet(packet, CommonCommand.READ_LCFG_RES)

    def set_discrete_fourier_transformation(self, dft_window: EDADFTWindow) -> Dict:
        """
        Set Discrete Fourier Transformation for EDA.

        :param dft_window: DFT window for Discrete Fourier Transformation, use get_supported_dft_windows()
                          | to list all supported DFT window.
        :type dft_window: EDADFTWindow
        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_eda_application()
            x = application.get_supported_dft_windows()
            print(x)
            # [<EDADFTWindow.DFT_WINDOW_4: ['0x0']>, ... ,<EDADFTWindow.DFT_WINDOW_32: ['0x3']>]
            x = application.set_discrete_fourier_transformation(application.DFT_WINDOW_32)
            print(x["payload"]["dft_window"])
            # EDADFTWindow.DFT_WINDOW_32
        """
        dft_window = self._dft_window_helper(dft_window)
        packet = DFTPacket(self._destination, EDACommand.SET_DFT_NUM_REQ)
        packet.set_dft_num(dft_window)
        return self._send_packet(packet, EDACommand.SET_DFT_NUM_RES)

    def write_device_configuration_block(self, addresses_values: List[List[int]]) -> Dict:
        """
        Writes the device configuration block values of specified addresses.
        This function takes a list of addresses and values to write, and returns a response packet as
        dictionary containing addresses and values.

        :param addresses_values: List of addresses and values to write.
        :type addresses_values: List[List[int]]
        :return: A response packet as dictionary.
        :rtype: Dict

        .. list-table::
           :widths: 50 50
           :header-rows: 1

           * - Address Lower Limit
             - Address Upper Limit
           * - 0x00
             - 0x02

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_eda_application()
            x = application.write_device_configuration_block([[0x0, 2], [0x1, 0x1]])
            print(x["payload"]["num_ops"])
            # 2

        """
        address_range = [0x00, 0x02]
        packet = DCBPacket(self._destination, DCBCommand.WRITE_CONFIG_REQ)
        num_ops = len(addresses_values)
        utils.check_array_address_range(addresses_values, address_range, 2)
        packet.set_size(utils.split_int_in_bytes(num_ops, length=2))
        packet.set_dcb_write_data(addresses_values, self._dcb_size, num_ops)
        return self._send_packet(packet, DCBCommand.WRITE_CONFIG_RES)

    def write_device_configuration_block_from_file(self, filename: str) -> Dict:
        """
        Writes the device configuration block values of specified addresses from file.

        :param filename: dcb filename
        :return: A response packet as dictionary.
        :rtype: Dict

        .. list-table::
           :widths: 50 50
           :header-rows: 1

           * - Address Lower Limit
             - Address Upper Limit
           * - 0x00
             - 0x02

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_eda_application()
            application.write_device_configuration_block_from_file("adxl_dcb.dcfg")

        """
        result = self._write_device_configuration_block_from_file_helper(filename)
        if result:
            return self.write_device_configuration_block(result)

    def write_dcb_to_lcfg(self) -> Dict:
        """
        Writes Device configuration block data to library configuration.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_eda_application()
            x = application.write_dcb_to_lcfg()
            print(x["payload"]["status"])
            # CommonStatus.OK
        """
        packet = CommandPacket(self._destination, CommonCommand.SET_LCFG_REQ)
        return self._send_packet(packet, CommonCommand.SET_LCFG_RES)

    def write_library_configuration(self, fields_values: List[List[int]]) -> Dict:
        """
        Writes library configuration from List of fields and values.

        :param fields_values: List of fields and values to write.
        :type fields_values: List[List[int]]
        :return: A response packet as dictionary.
        :rtype: Dict

        .. list-table::
           :widths: 50 50
           :header-rows: 1

           * - Fields Lower Limit
             - Fields Upper Limit
           * - 0x00
             - 0x02

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_eda_application()
            x = application.write_library_configuration([[0x00, 0x1]])
            print(x["payload"]["data"])
            # [['0x0', '0x1']]

        """
        address_range = [0x00, 0x02]
        packet = LibraryConfigReadWritePacket(self._destination, CommonCommand.WRITE_LCFG_REQ)
        fields_values_size = len(fields_values)
        packet.set_number_of_operations(fields_values_size)
        utils.check_array_address_range(fields_values, address_range, 2)
        packet.set_write_fields_data(fields_values)
        return self._send_packet(packet, CommonCommand.WRITE_LCFG_RES)

    def get_sensor_status(self) -> Dict:
        """
        Returns packet with number of subscribers and number of sensor start request registered.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_eda_application()
            x = application.get_sensor_status()
            print(x["payload"]["num_subscribers"], x["payload"]["num_start_registered"])
            # 0 0

        """
        return super().get_sensor_status()

    def set_callback(self, callback_function: Callable, args: Tuple = ()) -> None:
        """
        Sets the callback for the stream data.

        :param args: optional arguments that will be passed with the callback.
        :param callback_function: callback function for stream EDA data.
        :return: None

        .. code-block:: python3
            :emphasize-lines: 4,12

            from adi_study_watch import SDK

            # make sure optional arguments have default value to prevent them causing Exceptions.
            def callback(data, optional1=None, optional2=None):
                print(data)

            sdk = SDK("COM4")
            application = sdk.get_eda_application()
            # these optional arguments can be used to pass file, matplotlib or other objects to manipulate data.
            optional_arg1 = "1"
            optional_arg2 = "2"
            application.set_callback(callback, args=(optional_arg1, optional_arg2))

        """
        super().set_callback(callback_function, args)

    def set_timeout(self, timeout_value: float) -> None:
        """
        Sets the time out for queue to wait for command packet response.

        :param timeout_value: queue timeout value (in sec).
        :type timeout_value: int
        :return: None

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_eda_application()
            application.set_timeout(10)

        """
        super().set_timeout(timeout_value)

    def start_and_subscribe_stream(self) -> Tuple[Dict, Dict]:
        """
        Starts EDA sensor and also subscribe to the EDA stream.

        :return: A response packet as dictionary.
        :rtype: Tuple[Dict, Dict]

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_eda_application()
            start_sensor, subs_stream = application.start_and_subscribe_stream()
            print(start_sensor["payload"]["status"], subs_stream["payload"]["status"])
            # CommonStatus.STREAM_STARTED CommonStatus.SUBSCRIBER_ADDED
        """
        return super().start_and_subscribe_stream()

    def start_sensor(self) -> Dict:
        """
        Starts EDA sensor.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_eda_application()
            start_sensor = application.start_sensor()
            print(start_sensor["payload"]["status"])
            # CommonStatus.STREAM_STARTED
        """
        return super().start_sensor()

    def stop_and_unsubscribe_stream(self) -> Tuple[Dict, Dict]:
        """
        Stops EDA sensor and also Unsubscribe the EDA stream.

        :return: A response packet as dictionary.
        :rtype: Tuple[Dict, Dict]

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_eda_application()
            stop_sensor, unsubscribe_stream = application.stop_and_unsubscribe_stream()
            print(stop_sensor["payload"]["status"], unsubscribe_stream["payload"]["status"])
            # CommonStatus.STREAM_STOPPED CommonStatus.SUBSCRIBER_REMOVED
        """
        return super().stop_and_unsubscribe_stream()

    def stop_sensor(self) -> Dict:
        """
        Stops EDA sensor.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_eda_application()
            stop_sensor = application.stop_sensor()
            print(stop_sensor["payload"]["status"])
            # CommonStatus.STREAM_STOPPED
        """
        return super().stop_sensor()

    def subscribe_stream(self) -> Dict:
        """
        Subscribe to the EDA stream.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_eda_application()
            subs_stream = application.subscribe_stream()
            print(subs_stream["payload"]["status"])
            # CommonStatus.SUBSCRIBER_ADDED
        """
        return super().subscribe_stream()

    def unsubscribe_stream(self) -> Dict:
        """
        Unsubscribe the EDA stream.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_eda_application()
            unsubscribe_stream = application.unsubscribe_stream()
            print(unsubscribe_stream["payload"]["status"])
            # CommonStatus.SUBSCRIBER_REMOVED
        """
        return super().unsubscribe_stream()

    def _callback_data(self, packet, packet_id, callback_function=None, args=None):
        """
        Process and returns the data back to user's callback function.
        """
        self._callback_data_helper(packet, packet_id, EDADataPacket(), callback_function, args)

    def enable_csv_logging(self, filename, header=None) -> None:
        """
        Start logging stream data into CSV.

        :param filename: Name of the CSV file.
        :param header: Header list of the CSV file.
        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_eda_application()
            x = application.enable_csv_logging("eda.csv")
        """
        if header is None:
            header = ["Timestamp", "Seq No.", "Impedance Module", "Impedance Phase"]
        self._csv_logger[Stream.EDA] = CSVLogger(filename, header)

    def disable_csv_logging(self) -> None:
        """
        Stops logging stream data into CSV.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_eda_application()
            x = application.disable_csv_logging()
        """
        if self._csv_logger[Stream.EDA]:
            self._csv_logger[Stream.EDA].stop_logging()
        self._csv_logger[Stream.EDA] = None



# ******************************************************************************
# Copyright (c) 2019 Analog Devices, Inc.  All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# - Redistributions of source code must retain the above copyright notice, this
#  list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice,
#  this list of conditions and the following disclaimer in the documentation
#  and/or other materials provided with the distribution.
# - Modified versions of the software must be conspicuously marked as such.
# - This software is licensed solely and exclusively for use with
#  processors/products manufactured by or for Analog Devices, Inc.
# - This software may not be combined or merged with other code in any manner
#  that would cause the software to become subject to terms and conditions
#  which differ from those listed here.
# - Neither the name of Analog Devices, Inc. nor the names of its contributors
#  may be used to endorse or promote products derived from this software
#  without specific prior written permission.
# - The use of this software may or may not infringe the patent rights of one
#  or more patent holders.  This license does not release you from the
#  requirement that you obtain separate licenses from these patent holders to
#  use this software.
#
# THIS SOFTWARE IS PROVIDED BY ANALOG DEVICES, INC. AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# NONINFRINGEMENT, TITLE, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL ANALOG DEVICES, INC. OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, PUNITIVE OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, DAMAGES ARISING OUT OF
# CLAIMS OF INTELLECTUAL PROPERTY RIGHTS INFRINGEMENT; PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# ******************************************************************************

from .command_packet import CommandPacket
from ..enums.bcm_enums import HSResistorTIA


class DCBTimingInfoPacket(CommandPacket):

    def __init__(self, destination=None, command=None):
        super().__init__(destination, command)
        self._config["payload"]["clear_entries_time"] = {"size": 2, "join": True}
        self._config["payload"]["check_entries_time"] = {"size": 2, "join": True}
        self._config["payload"]["delete_record_time"] = {"size": 2, "join": True}
        self._config["payload"]["read_entry_time"] = {"size": 2, "join": True}
        self._config["payload"]["update_entry_time"] = {"size": 2, "join": True}


class FdsStatusPacket(CommandPacket):

    def __init__(self, destination=None, command=None):
        super().__init__(destination, command)
        self._config["payload"]["dirty_records"] = {"size": 2, "join": True}
        self._config["payload"]["open_records"] = {"size": 2, "join": True}
        self._config["payload"]["valid_records"] = {"size": 2, "join": True}
        self._config["payload"]["pages_available"] = {"size": 2, "join": True}
        self._config["payload"]["num_blocks"] = {"size": 2, "join": True}
        self._config["payload"]["blocks_free"] = {"size": 2, "join": True}


class HSRTIAPacket(CommandPacket):
    """
    Packet Structure:

    .. code-block::

        {
            'header': {
                'source': <Application.BCM: ['0xC3', '0x07']>,
                'destination': <Application.APP_USB: ['0xC7', '0x05']>,
                'length': '0xC',
                'checksum': '0x0'
            },
            'payload': {
                'command': <BCMCommand.SET_HS_TRANS_IMPEDANCE_AMPLIFIER_CAL_RES: ['0x49']>,
                'status': <CommonStatus.OK: ['0x00']>,
                'hs_resistor_tia': <HSResistorTIA.RESISTOR_1K: ['0x00', '0x01']>
            }
        }
    """

    def __init__(self, destination=None, command=None):
        super().__init__(destination, command)
        self._config["payload"]["hs_resistor_tia"] = {"size": 2}
        self._packet["payload"]["hs_resistor_tia"] = [0x00, 0x00]

    def set_hs_resistor_tia(self, hs_resistor_tia_id):
        self._packet["payload"]["hs_resistor_tia"] = hs_resistor_tia_id.value

    def decode_packet(self, data):
        super().decode_packet(data)
        self._packet["payload"]["hs_resistor_tia"] = HSResistorTIA(self._packet["payload"]["hs_resistor_tia"])



# ******************************************************************************
# Copyright (c) 2019 Analog Devices, Inc.  All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# - Redistributions of source code must retain the above copyright notice, this
#  list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice,
#  this list of conditions and the following disclaimer in the documentation
#  and/or other materials provided with the distribution.
# - Modified versions of the software must be conspicuously marked as such.
# - This software is licensed solely and exclusively for use with
#  processors/products manufactured by or for Analog Devices, Inc.
# - This software may not be combined or merged with other code in any manner
#  that would cause the software to become subject to terms and conditions
#  which differ from those listed here.
# - Neither the name of Analog Devices, Inc. nor the names of its contributors
#  may be used to endorse or promote products derived from this software
#  without specific prior written permission.
# - The use of this software may or may not infringe the patent rights of one
#  or more patent holders.  This license does not release you from the
#  requirement that you obtain separate licenses from these patent holders to
#  use this software.
#
# THIS SOFTWARE IS PROVIDED BY ANALOG DEVICES, INC. AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# NONINFRINGEMENT, TITLE, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL ANALOG DEVICES, INC. OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, PUNITIVE OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, DAMAGES ARISING OUT OF
# CLAIMS OF INTELLECTUAL PROPERTY RIGHTS INFRINGEMENT; PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# ******************************************************************************
import csv
import os
import sys
import logging
import threading
from typing import List
from datetime import datetime

import serial
import serial.tools.list_ports

from .core.packet_manager import PacketManager
from .core.enums.common_enums import CommonCommand
from .application.pm_application import PMApplication
from .application.fs_application import FSApplication
from .application.sqi_application import SQIApplication
from .application.ecg_application import ECGApplication
from .application.eda_application import EDAApplication
from .application.bcm_application import BCMApplication
from .application.ppg_application import PPGApplication
from .application.test_application import TestApplication
from .application.adxl_application import ADXLApplication
from .application.adpd_application import ADPDApplication
from .application.ad7156_application import AD7156Application
from .application.low_touch_application import LowTouchApplication
from .application.pedometer_application import PedometerApplication
from .application.temperature_application import TemperatureApplication

logger = logging.getLogger(__name__)


class SDK:
    """
    SDK class
    """

    def __init__(self, serial_port_address: str, is_ble: bool = False, baud_rate: int = 921600,
                 logging_filename: str = None, debug: bool = False, **kwargs):
        """
        Creates a SDK object

        :param serial_port_address: serial port of the device connected
        :param baud_rate: baud rate
        :param **kwargs: other keys and values  for creating serial object
        """
        self._serial_object = None
        self._packet_manager = None
        self.connect(serial_port_address, is_ble, baud_rate, logging_filename, debug, **kwargs)

    def connect(self, serial_port_address: str, is_ble: bool = False, baud_rate: int = 921600,
                logging_filename: str = None, debug: bool = False, **kwargs):
        """
        Connect method allows you to reconnect to SDK; you must call disconnect before using connect.
        """

        log_format = '%(asctime)s - %(threadName)s - %(name)s - %(levelname)s - %(message)s'
        date_format = '%Y-%m-%d %H:%M:%S'
        if logging_filename:
            logging.basicConfig(filename=logging_filename, filemode='a', format=log_format, datefmt=date_format,
                                level=logging.DEBUG)
        else:
            if debug:
                logging.basicConfig(format=log_format, datefmt=date_format, level=logging.DEBUG)
            else:
                logging.basicConfig(format=log_format, datefmt=date_format)
        logger.debug("----- Study Watch SDK Started -----")
        self._serial_object = serial.Serial(serial_port_address, baud_rate, **kwargs)
        self._packet_manager = PacketManager(self._serial_object)
        if is_ble:
            self._packet_manager.set_ble_source()
        else:
            self._packet_manager.set_usb_source()

        self._packet_manager.start_receive_and_process_threads()
        self.get_pm_application().set_datetime(datetime.now())

    def get_adpd_application(self, callback_function_default=None, args=()):
        """
        Creates an adpd application object

        :param callback_function_default: callback function for all adpd slot stream.
        :type callback_function_default: Callable
        :param args: optional arguments that will be passed back with the callback.
        :returns: an Adpd Application
        :rtype: ADPDApplication
        """
        return ADPDApplication(callback_function_default, self._packet_manager, args)

    def get_adxl_application(self, callback_function=None, args=()):
        """
        Creates an adxl application object

        :param callback_function: callback function for adxl stream
        :type callback_function: Callable
        :param args: optional arguments that will be passed back with the callback.
        :returns: an Adxl Application
        :rtype: ADXLApplication
        """
        return ADXLApplication(callback_function, self._packet_manager, args)

    def get_ecg_application(self, callback_function=None, args=()):
        """
        Creates an ecg application object

        :param callback_function: callback function for ecg stream
        :type callback_function: Callable
        :param args: optional arguments that will be passed back with the callback.
        :returns: an ecg Application
        :rtype: ECGApplication
        """
        return ECGApplication(callback_function, self._packet_manager, args)

    def get_eda_application(self, callback_function=None, args=()):
        """
        Creates an eda application object

        :param callback_function: callback function for eda stream
        :type callback_function: Callable
        :param args: optional arguments that will be passed back with the callback.
        :returns: an eda Application
        :rtype: EDAApplication
        """
        return EDAApplication(callback_function, self._packet_manager, args)

    def get_fs_application(self):
        """
        Creates an fs application object

        :returns: an fs Application
        :rtype: FSApplication
        """
        return FSApplication(self._packet_manager)

    def get_pedometer_application(self, callback_function=None, args=()):
        """
        Creates an pedometer application object

        :param callback_function: callback function for pedometer stream
        :type callback_function: Callable
        :param args: optional arguments that will be passed back with the callback.
        :returns: an pedometer Application
        :rtype: PedometerApplication
        """
        return PedometerApplication(callback_function, self._packet_manager, args)

    def get_pm_application(self):
        """
        Creates an pm application object

        :returns: an pm Application
        :rtype: PMApplication
        """
        return PMApplication(self._packet_manager)

    def get_ppg_application(self, callback_ppg=None, callback_syncppg=None, args_ppg=(), args_syncppg=()):
        """
        Creates an ppg application object

        :param callback_ppg: callback function for ppg stream.
        :type callback_ppg: Callable
        :param callback_syncppg: callback function for sync ppg stream.
        :type callback_syncppg: Callable
        :param args_ppg: optional arguments that will be passed back with ppg callback.
        :param args_syncppg: optional arguments that will be passed back with sync ppg callback.
        :returns: an Ppg Application
        :rtype: PPGApplication
        """
        return PPGApplication(callback_ppg, callback_syncppg, self._packet_manager, args_ppg, args_syncppg)

    def get_temperature_application(self, callback_function=None, args=()):
        """
        Creates a temperature application object

        :param callback_function: callback function for temperature stream
        :type callback_function: Callable
        :param args: optional arguments that will be passed back with the callback.
        :returns: a Temperature Application
        :rtype: TemperatureApplication
        """
        return TemperatureApplication(callback_function, self._packet_manager, args)

    def get_sqi_application(self, callback_function=None, args=()):
        """
        Creates a sqi application object

        :param callback_function: callback function for sqi stream
        :type callback_function: Callable
        :param args: optional arguments that will be passed back with the callback.
        :returns: a SQI Application
        :rtype: SQIApplication
        """
        return SQIApplication(callback_function, self._packet_manager, args)

    def get_bcm_application(self, callback_function=None, args=()):
        """
        Creates a bcm application object

        :param callback_function: callback function for bcm stream
        :type callback_function: Callable
        :param args: optional arguments that will be passed back with the callback.
        :returns: a BCM Application
        :rtype: BCMApplication
        """
        return BCMApplication(callback_function, self._packet_manager, args)

    def get_ad7156_application(self):
        """
        Creates a ad7156 application object

        :returns: a AD7156 Application
        :rtype: AD7156Application
        """
        return AD7156Application(self._packet_manager)

    def get_low_touch_application(self):
        """
        Creates a low touch application object

        :returns: a LowTouch Application
        :rtype: LowTouchApplication
        """
        return LowTouchApplication(self._packet_manager)

    def get_test_application(self, key_test_callback=None, cap_sense_callback=None):
        """
        Creates a test application object, used for internal firmware testing.

        :returns: a Test Application
        :rtype: TestApplication
        """
        return TestApplication(key_test_callback, cap_sense_callback, self._packet_manager)

    def unsubscribe_all_streams(self):
        """
        Unsubscribe from all application streams
        """
        result = [self.get_adxl_application().unsubscribe_stream(), self.get_sqi_application().unsubscribe_stream(),
                  self.get_ppg_application().unsubscribe_stream(), self.get_bcm_application().unsubscribe_stream(),
                  self.get_ecg_application().unsubscribe_stream(), self.get_eda_application().unsubscribe_stream(),
                  self.get_temperature_application().unsubscribe_stream(),
                  self.get_pedometer_application().unsubscribe_stream()]
        adpd_app = self.get_adpd_application()
        fs_app = self.get_fs_application()
        for stream in adpd_app.get_supported_streams():
            result.append(adpd_app.unsubscribe_stream(stream))
        for stream in fs_app.get_supported_streams():
            result.append(fs_app.unsubscribe_stream(stream))
        return result

    @staticmethod
    def get_available_ports() -> List:
        """
        returns the list of tuple (port, description, hardware_id) of available ports.
        """
        ports = serial.tools.list_ports.comports()
        result = []
        for port, desc, hardware_id in sorted(ports):
            result.append((port, desc, hardware_id))
        return result

    @staticmethod
    def join_csv(*args, output_filename="combined.csv"):
        """
        Joins multiple data stream csv file into single csv file.
        """
        rows = {}
        space = {}
        max_lines = 0
        header = None
        # reading each file
        for file in args:
            try:
                csv_file = open(file, 'r')
                reader = csv.reader(csv_file, quoting=csv.QUOTE_NONE, delimiter=',')
                csv_rows = []
                max_space = 0
                # reading each row
                for row in reader:
                    if len(row) > 0:
                        csv_rows.append(row)
                    # max length of row in a file, used in the case where files have unequal sizes
                    max_space = max(max_space, len(row))
                space[file] = max_space
                rows[file] = csv_rows
                # obtaining header
                if header is None:
                    header = rows[file][:2]
                rows[file] = rows[file][2:]
                # max lines in all files
                max_lines = max(max_lines, len(rows[file]))
                csv_file.close()
            except Exception as e:
                logger.debug(e)

        # creating required rows in combined csv
        final_rows = [[]] * max_lines
        for file in rows:
            for i in range(max_lines):
                # adding spaces because no data, check for final_rows[i] += [" "] * space[file]
                if i + 1 > len(rows[file]):
                    final_rows[i] = final_rows[i] + [" "] * space[file]
                else:
                    # appending data to the row
                    final_rows[i] = final_rows[i] + rows[file][i]

        # writing data to final csv
        final_rows = header + final_rows
        file = open(output_filename, 'w', newline="")
        writer = csv.writer(file, quoting=csv.QUOTE_NONE, escapechar='\'')
        for row in final_rows:
            writer.writerow(row)
        file.close()

    @staticmethod
    def _sub_helper(app, stream, packet_manager):
        packet_id = app._get_packet_id(CommonCommand.STREAM_DATA, app._stream if stream is None else stream)
        if type(app) == ADPDApplication:
            packet_manager.subscribe(packet_id, app._stream_to_callback[stream])
        else:
            packet_manager.subscribe(packet_id, app._callback_data)

    @staticmethod
    def convert_log_to_csv(filename):
        """
        Converts M2M2 log file into csv.
        """
        folder_name = filename.split(".")[0]
        try:
            os.mkdir(folder_name)
        except Exception as e:
            logger.debug(e)
        # creating all applications
        packet_manager = PacketManager(None, filename=filename)
        packet_manager.set_usb_source()
        adpd_app = ADPDApplication(None, packet_manager, None)
        adxl_app = ADXLApplication(None, packet_manager, None)
        ecg_app = ECGApplication(None, packet_manager, None)
        eda_app = EDAApplication(None, packet_manager, None)
        ped_app = PedometerApplication(None, packet_manager, None)
        ppg_app = PPGApplication(None, None, packet_manager, None, None)
        temp_app = TemperatureApplication(None, packet_manager, None)
        sqi_app = SQIApplication(None, packet_manager, None)
        bcm_app = BCMApplication(None, packet_manager, None)
        pm_app = PMApplication(packet_manager)

        # enabling csv logging
        adxl_app.enable_csv_logging(f"{folder_name}/adxl.csv")
        for i, stream in enumerate(adpd_app.get_supported_streams()):
            adpd_app.enable_csv_logging(f"{folder_name}/adpd{i + 1}.csv", stream=stream)
        ecg_app.enable_csv_logging(f"{folder_name}/ecg.csv")
        eda_app.enable_csv_logging(f"{folder_name}/eda.csv")
        ped_app.enable_csv_logging(f"{folder_name}/ped.csv")
        ppg_app.enable_csv_logging(f"{folder_name}/ppg.csv", stream=ppg_app.PPG)
        ppg_app.enable_csv_logging(f"{folder_name}/sync_ppg.csv", stream=ppg_app.SYNC_PPG)
        temp_app.enable_csv_logging(f"{folder_name}/temp.csv")
        sqi_app.enable_csv_logging(f"{folder_name}/sqi.csv")
        bcm_app.enable_csv_logging(f"{folder_name}/bcm.csv")

        # subscribing
        apps = [adxl_app, ecg_app, eda_app, ped_app, ppg_app, temp_app, sqi_app, bcm_app]
        for app in apps:
            SDK._sub_helper(app, None, packet_manager)
        for stream in adpd_app.get_supported_streams():
            SDK._sub_helper(adpd_app, stream, packet_manager)

        info_result = {}

        # callback for sum initial packets
        def info_packets(function, key):
            if key == "ecg_lcfg":
                response_packet = function([0x0])
            else:
                response_packet = function()
            info_result[key] = response_packet
            if key == "datetime":
                dt = datetime(response_packet['payload']['year'], response_packet['payload']['month'],
                              response_packet['payload']['day'], response_packet['payload']['hour'],
                              response_packet['payload']['minute'], response_packet['payload']['second'])
                ticks = (32000.0 * ((response_packet['payload']['hour'] * 3600) +
                                    (response_packet['payload']['minute'] * 60) + response_packet['payload']['second']))
                ref_timestamp = datetime.timestamp(dt)
                last_timestamp = [ref_timestamp, ticks]
                for _stream in adpd_app.get_supported_streams():
                    adpd_app._last_timestamp_adpd[_stream] = last_timestamp
                adxl_app._last_timestamp = last_timestamp
                ecg_app._last_timestamp = last_timestamp
                eda_app._last_timestamp = last_timestamp
                ped_app._last_timestamp = last_timestamp
                ppg_app._last_timestamp_ppg = last_timestamp
                ppg_app._last_timestamp_syncppg = last_timestamp
                temp_app._last_timestamp = last_timestamp
                sqi_app._last_timestamp = last_timestamp
                bcm_app._last_timestamp = last_timestamp

        # threads for initial packets
        threading.Thread(target=info_packets, args=(pm_app.get_datetime, "datetime")).start()
        threading.Thread(target=info_packets, args=(pm_app.get_system_info, "system_info")).start()
        threading.Thread(target=info_packets, args=(pm_app.get_version, "version")).start()
        threading.Thread(target=info_packets, args=(ppg_app.get_algo_version, "ppg_algo_version")).start()
        threading.Thread(target=info_packets, args=(ped_app.get_algo_version, "ped_algo_version")).start()
        threading.Thread(target=info_packets, args=(ecg_app.get_algo_version, "ecg_algo_version")).start()
        threading.Thread(target=info_packets, args=(sqi_app.get_algo_version, "sqi_algo_version")).start()
        threading.Thread(target=info_packets, args=(adpd_app.get_device_configuration, "adpd_dcfg")).start()
        threading.Thread(target=info_packets, args=(adxl_app.get_device_configuration, "adxl_dcfg")).start()
        threading.Thread(target=info_packets, args=(ppg_app.get_library_configuration, "ppg_lcfg")).start()
        threading.Thread(target=info_packets, args=(ecg_app.read_library_configuration, "ecg_lcfg")).start()

        packet_manager.start_receive_and_process_threads()

        # waiting for packet manager to finish processing log file.
        while not packet_manager.queue.empty() or packet_manager.processing_file:
            pass

        # disabling csv logging.
        adxl_app.disable_csv_logging()
        for stream in adpd_app.get_supported_streams():
            adpd_app.disable_csv_logging(stream=stream)
        ecg_app.disable_csv_logging()
        eda_app.disable_csv_logging()
        ped_app.disable_csv_logging()
        ppg_app.disable_csv_logging(stream=ppg_app.PPG)
        ppg_app.disable_csv_logging(stream=ppg_app.SYNC_PPG)
        temp_app.disable_csv_logging()
        sqi_app.disable_csv_logging()
        bcm_app.disable_csv_logging()

        # storing initial packets in info.csv
        with open(f"{folder_name}/info.csv", 'w') as f:
            for res in info_result:
                f.write(res + " :: " + str(info_result[res]) + "\n")

        # combining adpd csv
        adpd_csv_files = []
        for i in range(1, 13):
            adpd_csv_files.append(f"{folder_name}/adpd{i}.csv")
        SDK.join_csv(*adpd_csv_files, output_filename=f"{folder_name}/adpd_streams.csv")

    def disconnect(self):
        """disconnect SDK"""
        logger.debug("----- Study Watch SDK Stopped -----")
        self._packet_manager.close()



import csv
import math
import logging
import time
from datetime import datetime

from ..core.enums.common_enums import Stream

logger = logging.getLogger(__name__)


class CSVLogger:
    def __init__(self, filename, header, write_header=True):
        self.ch1 = []
        self.ch2 = []
        self.file = None
        self.ch1_dark = []
        self.ch2_dark = []
        self.ch1_time = []
        self.ch2_time = []
        self.writer = None
        self.header = header
        self.filename = filename
        self.logging_started = None
        self.write_header = write_header

    def write_row(self, row):
        if self.writer:
            self.writer.writerow(row)

    def start_logging(self, last_timestamp):
        try:
            self.file = open(self.filename, 'w', newline="")
            self.writer = csv.writer(self.file, quoting=csv.QUOTE_NONE)
            dt_object = datetime.fromtimestamp(last_timestamp)
            is_dst = time.daylight and time.localtime().tm_isdst > 0
            utc_offset = - (time.altzone if is_dst else time.timezone)
            self.write_row(["Local: ", dt_object, "tz_sec ", utc_offset])
            self.write_row(["Linux Epoch (ms):", last_timestamp])
            if self.write_header:
                self.write_row(self.header)
        except Exception as e:
            logger.error(f"Error while opening the {self.filename} file, reason :: {e}.", exc_info=True)

    def stop_logging(self):
        data_len = max(len(self.ch1_time), len(self.ch2_time))
        if len(self.ch1_time) > 0 and len(self.ch2_time) > 0:
            if not len(self.ch1_time) == len(self.ch2_time):
                data_len = min(len(self.ch1_time), len(self.ch2_time))
                logger.warning(f"Issue with ADPD CSV logging. Recommended sequence unsubscribe_stream "
                               f"then disable_csv_logging.")
        for i in range(data_len):
            if len(self.ch2) == 0:
                self.adpd_add_row([self.ch1_time[i], self.ch1_dark[i], self.ch1[i]], None)
            elif len(self.ch1) == 0:
                self.adpd_add_row([self.ch2_time[i], self.ch2_dark[i], self.ch2[i]], None)
            else:
                self.adpd_add_row(
                    [self.ch1_time[i], self.ch1_dark[i], self.ch1[i], self.ch2_dark[i], self.ch2[i]], None)

        if self.file:
            self.file.close()

    def add_row(self, result, last_timestamp):
        stream = result["header"]["source"]
        if self.logging_started is None:
            self.logging_started = True
            self.start_logging(last_timestamp)
        if stream == Stream.ADXL:
            self._adxl_callback(result)
        elif stream in [Stream.ADPD1, Stream.ADPD2, Stream.ADPD3, Stream.ADPD4, Stream.ADPD5, Stream.ADPD6,
                        Stream.ADPD7, Stream.ADPD8, Stream.ADPD9, Stream.ADPD10, Stream.ADPD11, Stream.ADPD12]:
            self._adpd_callback(result)
        elif stream == Stream.PPG:
            self._ppg_callback(result)
        elif stream == Stream.SYNC_PPG:
            self._syncppg_callback(result)
        elif stream == Stream.ECG:
            self._ecg_callback(result)
        elif stream == Stream.EDA:
            self._eda_callback(result)
        elif stream == Stream.TEMPERATURE:
            self._temp_callback(result)
        elif stream == Stream.PEDOMETER:
            self._ped_callback(result)
        elif stream == Stream.BCM:
            self._bcm_callback(result)
        elif stream == Stream.SQI:
            self._sqi_callback(result)

    def adpd_add_row(self, row, header_format):
        if not self.write_header:
            self.write_header = True
            row1 = self.header[0]
            self.write_row([" "] + [row1, "", "", "", ""])
            row2 = []
            if header_format == 0:
                row2 = ["CH1", "", "", "CH2", ""]
            elif header_format == 1:
                row2 = ["CH1", "", ""]
            elif header_format == 2:
                row2 = ["CH2", "", ""]
            self.write_row([" "] + row2)
            row3 = []
            if header_format == 0:
                row3 = ["Timestamp", "D1", "S1", "D2", "S2"]
            elif header_format == 1:
                row3 = ["Timestamp", "D1", "S1"]
            elif header_format == 2:
                row3 = ["Timestamp", "D2", "S2"]
            self.write_row([" "] + row3)
        self.write_row([" "] + row)

    def _adxl_callback(self, data):
        for value in data["payload"]["stream_data"]:
            self.write_row([value["timestamp"], value["x"], value["y"], value["z"]])

    def _adpd_callback(self, data):
        if len(self.ch1_time) > 200 or len(self.ch2_time) > 200:
            adpd_data_length = len(data["payload"]["signal_data"])
            for i in range(adpd_data_length):
                if len(self.ch2) == 0:
                    self.adpd_add_row([self.ch1_time[i], self.ch1_dark[i], self.ch1[i]], 1)
                elif len(self.ch1) == 0:
                    self.adpd_add_row([self.ch2_time[i], self.ch2_dark[i], self.ch2[i]], 2)
                else:
                    self.adpd_add_row(
                        [self.ch1_time[i], self.ch1_dark[i], self.ch1[i], self.ch2_dark[i], self.ch2[i]], 0)

            del self.ch1[:adpd_data_length]
            del self.ch2[:adpd_data_length]
            del self.ch1_dark[:adpd_data_length]
            del self.ch2_dark[:adpd_data_length]
            del self.ch1_time[:adpd_data_length]
            del self.ch2_time[:adpd_data_length]

        if data["payload"]["channel_num"] == 1:
            for signal_data in data["payload"]["signal_data"]:
                self.ch1_time.append(data["payload"]["timestamp"])
                self.ch1.append(signal_data)
            for dark_data in data["payload"]["dark_data"]:
                self.ch1_dark.append(dark_data)

        elif data["payload"]["channel_num"] == 2:
            for adpd_data in data["payload"]["signal_data"]:
                self.ch2_time.append(data["payload"]["timestamp"])
                self.ch2.append(adpd_data)
            for dark_data in data["payload"]["dark_data"]:
                self.ch2_dark.append(dark_data)

    def _ecg_callback(self, data):
        for value in data["payload"]["stream_data"]:
            self.write_row([value["timestamp"], data["payload"]["sequence_number"], value["ecg_data"]])

    def _ppg_callback(self, data):
        hr = float(data["payload"]["hr"]) / float(16)
        confidence = (float(100) * (data["payload"]["confidence"] / float(1024)))
        self.write_row([data["payload"]["timestamp"], hr, confidence, data["payload"]["hr_type"]])

    def _syncppg_callback(self, data):
        for value in data["payload"]["stream_data"]:
            self.write_row(
                [value["ppg_timestamp"], value["ppg_data"], value["adxl_timestamp"], value["adxl_x"],
                 value["adxl_y"], value["adxl_z"]])

    def _eda_callback(self, data):
        for value in data["payload"]["stream_data"]:
            eda_real = value["real_data"]
            eda_imaginary = value["imaginary_data"]
            if eda_real == 0:
                eda_real = 1
            impedance_img = eda_imaginary * 1000
            impedance_real = eda_real * 1000
            real_and_img = float(impedance_real * impedance_real + impedance_img * impedance_img)
            impedance_module = math.sqrt(real_and_img)
            impedance_phase = math.atan2(impedance_img, impedance_real)
            self.write_row([value["timestamp"], data["payload"]["sequence_number"],
                            impedance_module, impedance_phase])

    def _temp_callback(self, data):
        self.write_row([data["payload"]["timestamp"], data["payload"]["skin_temperature"],
                        data["payload"]["impedance"]])

    def _ped_callback(self, data):
        self.write_row([data["payload"]["timestamp"], data["payload"]["steps"]])

    def _bcm_callback(self, data):
        for value in data["payload"]["stream_data"]:
            bcm_real = value["real"]
            bcm_imaginary = value["imaginary"]
            if bcm_real == 0:
                bcm_real = 1
            if bcm_imaginary == 0:
                bcm_imaginary = 1
            impedance_img = bcm_imaginary
            impedance_real = bcm_real
            real_and_img = float(impedance_real * impedance_real + impedance_img * impedance_img)
            impedance_magnitude = math.sqrt(real_and_img)
            impedance_phase = math.atan2(impedance_img, impedance_real)
            self.write_row([value["timestamp"], data["payload"]["sequence_number"],
                            impedance_magnitude, impedance_phase])

    def _sqi_callback(self, data):
        self.write_row([data["payload"]["timestamp"], data["payload"]["sqi"]])



# ******************************************************************************
# Copyright (c) 2019 Analog Devices, Inc.  All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# - Redistributions of source code must retain the above copyright notice, this
#  list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice,
#  this list of conditions and the following disclaimer in the documentation
#  and/or other materials provided with the distribution.
# - Modified versions of the software must be conspicuously marked as such.
# - This software is licensed solely and exclusively for use with
#  processors/products manufactured by or for Analog Devices, Inc.
# - This software may not be combined or merged with other code in any manner
#  that would cause the software to become subject to terms and conditions
#  which differ from those listed here.
# - Neither the name of Analog Devices, Inc. nor the names of its contributors
#  may be used to endorse or promote products derived from this software
#  without specific prior written permission.
# - The use of this software may or may not infringe the patent rights of one
#  or more patent holders.  This license does not release you from the
#  requirement that you obtain separate licenses from these patent holders to
#  use this software.
#
# THIS SOFTWARE IS PROVIDED BY ANALOG DEVICES, INC. AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# NONINFRINGEMENT, TITLE, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL ANALOG DEVICES, INC. OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, PUNITIVE OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, DAMAGES ARISING OUT OF
# CLAIMS OF INTELLECTUAL PROPERTY RIGHTS INFRINGEMENT; PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# ******************************************************************************

import struct
from .. import utils
from .command_packet import CommandPacket


class ADPDDataPacket(CommandPacket):
    """
    Packet Structure:

    .. code-block::

        {
            'header': {
                'source': <Stream.ADPD6: ['0xC2', '0x16']>,
                'destination': <Application.APP_USB: ['0xC7', '0x05']>,
                'length': '0x2C',
                'checksum': '0x0'
            },
            'payload': {
                'command': <CommonCommand.STREAM_DATA: ['0x28']>,
                'status': <CommonStatus.OK: ['0x00']>,
                'sequence_number': 0,
                'data_type': 4,
                'channel_num': 2,
                'timestamp': 1751731879,
                'sample_num': 2,
                'signal_data': [ 351762, 351755 ],
                'dark_data': [ 0, 0 ],
            }
        }
    """

    def __init__(self):
        super().__init__()
        self._config["payload"]["sequence_number"] = {"size": 2, "join": True}
        self._config["payload"]["data_type"] = {"size": 2, "join": True}
        self._config["payload"]["channel_num"] = {"size": 1, "join": True}
        self._config["payload"]["timestamp"] = {"size": 4, "join": True}
        self._config["payload"]["sample_num"] = {"size": 1, "join": True}
        self._config["payload"]["data"] = {"size": 25}

    def decode_packet(self, data):
        super().decode_packet(data)
        self._packet["payload"]["dark_data"] = []
        self._packet["payload"]["signal_data"] = []
        if self._packet["payload"]["data_type"] & 0x100 == 0:
            signal_size = (self._packet["payload"]["data_type"] & 0xF)
            dark_size = (self._packet["payload"]["data_type"] & 0xF0) >> 4
            data_size = signal_size + dark_size
            sample_num = self._packet["payload"]["sample_num"]
            for i in range(0, sample_num * data_size, data_size):
                data = self._packet["payload"]["data"][i: i + dark_size]
                dark_data = utils.join_multi_length_packets(data)
                data = self._packet["payload"]["data"][i + dark_size: i + dark_size + signal_size]
                signal_data = utils.join_multi_length_packets(data)
                self._packet["payload"]["dark_data"].append(dark_data)
                self._packet["payload"]["signal_data"].append(signal_data)
        else:
            signal_size = (self._packet["payload"]["data_type"] & 0xFF)
            for i in range(0, signal_size, 2):
                self._packet["payload"]["dark_data"].append(0)
                data = self._packet["payload"]["data"][i: i + 1]
                signal_data = utils.join_multi_length_packets(data)
                self._packet["payload"]["signal_data"].append(signal_data)
        del self._packet["payload"]["data"]

    def update_timestamp(self, last_timestamp):
        reference_time, last_ts = last_timestamp
        timestamp = self._packet["payload"]["timestamp"]
        reference_time = utils.get_updated_timestamp(reference_time, last_ts, timestamp)
        self._packet["payload"]["timestamp"] = reference_time * 1000
        last_timestamp[0] = reference_time
        last_timestamp[1] = timestamp


class ADXLDataPacket(CommandPacket):
    """
    Packet Structure:

    .. code-block::

        {
            'header': {
                'source': <Stream.ADXL: ['0xC2', '0x02']>,
                'destination': <Application.APP_USB: ['0xC7', '0x05']>,
                'length': '0x37',
                'checksum': '0x0'
            },
            'payload': {
                'command': <CommonCommand.STREAM_DATA: ['0x28']>,
                'status': <CommonStatus.OK: ['0x00']>,
                'sequence_number': 395,
                'data_type': 0,
                'stream_data': [
                    {
                        'timestamp': 1767361372,
                        'x': -85,
                        'y': 90,
                        'z': 55
                    },
                    {
                        'timestamp': 1767362027,
                        'x': -80,
                        'y': 88,
                        'z': 55
                    },
                    {
                        'timestamp': 1767362682,
                        'x': 40,
                        'y': 274,
                        'z': 79
                    },
                    {
                        'timestamp': 1767363337,
                        'x': 70,
                        'y': 273,
                        'z': 54
                    },
                    {
                        'timestamp': 1767363931,
                        'x': 57,
                        'y': 257,
                        'z': 48
                    }
                ]
            }
        }
    """

    def __init__(self):
        super().__init__()
        self._config["payload"]["sequence_number"] = {"size": 2, "join": True}
        self._config["payload"]["data_type"] = {"size": 1, "join": True}
        self._config["payload"]["timestamp1"] = {"size": 4, "join": True}
        self._config["payload"]["x1"] = {"size": 2, "join": True, "signed": True}
        self._config["payload"]["y1"] = {"size": 2, "join": True, "signed": True}
        self._config["payload"]["z1"] = {"size": 2, "join": True, "signed": True}
        self._config["payload"]["timestamp2"] = {"size": 2, "join": True}
        self._config["payload"]["x2"] = {"size": 2, "join": True, "signed": True}
        self._config["payload"]["y2"] = {"size": 2, "join": True, "signed": True}
        self._config["payload"]["z2"] = {"size": 2, "join": True, "signed": True}
        self._config["payload"]["timestamp3"] = {"size": 2, "join": True}
        self._config["payload"]["x3"] = {"size": 2, "join": True, "signed": True}
        self._config["payload"]["y3"] = {"size": 2, "join": True, "signed": True}
        self._config["payload"]["z3"] = {"size": 2, "join": True, "signed": True}
        self._config["payload"]["timestamp4"] = {"size": 2, "join": True}
        self._config["payload"]["x4"] = {"size": 2, "join": True, "signed": True}
        self._config["payload"]["y4"] = {"size": 2, "join": True, "signed": True}
        self._config["payload"]["z4"] = {"size": 2, "join": True, "signed": True}
        self._config["payload"]["timestamp5"] = {"size": 2, "join": True}
        self._config["payload"]["x5"] = {"size": 2, "join": True, "signed": True}
        self._config["payload"]["y5"] = {"size": 2, "join": True, "signed": True}
        self._config["payload"]["z5"] = {"size": 2, "join": True, "signed": True}

    def decode_packet(self, data):
        super().decode_packet(data)
        self._packet["payload"]["stream_data"] = []
        timestamp = 0
        for i in range(1, 6):
            timestamp += self._packet["payload"][f"timestamp{i}"]
            data = {"timestamp": timestamp,
                    "x": self._packet["payload"][f"x{i}"],
                    "y": self._packet["payload"][f"y{i}"],
                    "z": self._packet["payload"][f"z{i}"]}
            [self._packet["payload"].pop(key) for key in [f"timestamp{i}", f"x{i}", f"y{i}", f"z{i}"]]
            self._packet["payload"]["stream_data"].append(data)

    def update_timestamp(self, last_timestamp):
        for i in range(len(self._packet["payload"]["stream_data"])):
            reference_time, last_ts = last_timestamp
            timestamp = self._packet["payload"]["stream_data"][i]["timestamp"]
            reference_time = utils.get_updated_timestamp(reference_time, last_ts, timestamp)
            self._packet["payload"]["stream_data"][i]["timestamp"] = reference_time * 1000
            last_timestamp[0] = reference_time
            last_timestamp[1] = timestamp


class ECGDataPacket(CommandPacket):
    """
    Packet Structure:

    .. code-block::

        {
            'header': {
                'source': <Stream.ECG: ['0xC4', '0x01']>,
                'destination': <Application.APP_USB: ['0xC7', '0x05']>,
                'length': '0x3D',
                'checksum': '0x0'
            },
            'payload': {
                'command': <CommonCommand.STREAM_DATA: ['0x28']>,
                'status': <CommonStatus.OK: ['0x00']>,
                'sequence_number': 0,
                'data_type': 1,
                'ecg_info': 0,
                'hr': 0,
                'stream_data': [
                    {
                        'timestamp': 1771095376,
                        'ecg_data': 12351
                    },
                    {
                        'timestamp': 1771095703,
                        'ecg_data': 12353
                    },
                    {
                        'timestamp': 1771096030,
                        'ecg_data': 52470
                    },
                    {
                        'timestamp': 1771096357,
                        'ecg_data': 41129
                    },
                    {
                        'timestamp': 1771096676,
                        'ecg_data': 63838
                    },
                    {
                        'timestamp': 1771096995,
                        'ecg_data': 63848
                    },
                    {
                        'timestamp': 1771097314,
                        'ecg_data': 63848
                    },
                    {
                        'timestamp': 1771097633,
                        'ecg_data': 63848
                    },
                    {
                        'timestamp': 1771097954,
                        'ecg_data': 63833
                    },
                    {
                        'timestamp': 1771098273,
                        'ecg_data': 63846
                    },
                    {
                        'timestamp': 1771098592,
                        'ecg_data': 63846
                    }
                ]
            }
        }
    """

    def __init__(self):
        super().__init__()
        self._config["payload"]["sequence_number"] = {"size": 2, "join": True}
        self._config["payload"]["data_type"] = {"size": 1, "join": True}
        self._config["payload"]["timestamp1"] = {"size": 4, "join": True}
        self._config["payload"]["ecg_info"] = {"size": 1, "join": True}
        self._config["payload"]["hr"] = {"size": 1, "join": True}
        self._config["payload"]["ecg_data1"] = {"size": 2, "join": True}
        self._config["payload"]["timestamp2"] = {"size": 2, "join": True}
        self._config["payload"]["ecg_data2"] = {"size": 2, "join": True}
        self._config["payload"]["timestamp3"] = {"size": 2, "join": True}
        self._config["payload"]["ecg_data3"] = {"size": 2, "join": True}
        self._config["payload"]["timestamp4"] = {"size": 2, "join": True}
        self._config["payload"]["ecg_data4"] = {"size": 2, "join": True}
        self._config["payload"]["timestamp5"] = {"size": 2, "join": True}
        self._config["payload"]["ecg_data5"] = {"size": 2, "join": True}
        self._config["payload"]["timestamp6"] = {"size": 2, "join": True}
        self._config["payload"]["ecg_data6"] = {"size": 2, "join": True}
        self._config["payload"]["timestamp7"] = {"size": 2, "join": True}
        self._config["payload"]["ecg_data7"] = {"size": 2, "join": True}
        self._config["payload"]["timestamp8"] = {"size": 2, "join": True}
        self._config["payload"]["ecg_data8"] = {"size": 2, "join": True}
        self._config["payload"]["timestamp9"] = {"size": 2, "join": True}
        self._config["payload"]["ecg_data9"] = {"size": 2, "join": True}
        self._config["payload"]["timestamp10"] = {"size": 2, "join": True}
        self._config["payload"]["ecg_data10"] = {"size": 2, "join": True}
        self._config["payload"]["timestamp11"] = {"size": 2, "join": True}
        self._config["payload"]["ecg_data11"] = {"size": 2, "join": True}

    def decode_packet(self, data):
        super().decode_packet(data)
        self._packet["payload"]["stream_data"] = []
        timestamp = 0
        for i in range(1, 12):
            timestamp += self._packet["payload"][f"timestamp{i}"]
            data = {"timestamp": timestamp,
                    "ecg_data": self._packet["payload"][f"ecg_data{i}"]}
            [self._packet["payload"].pop(key) for key in [f"timestamp{i}", f"ecg_data{i}"]]
            self._packet["payload"]["stream_data"].append(data)

    def update_timestamp(self, last_timestamp):
        for i in range(len(self._packet["payload"]["stream_data"])):
            reference_time, last_ts = last_timestamp
            timestamp = self._packet["payload"]["stream_data"][i]["timestamp"]
            reference_time = utils.get_updated_timestamp(reference_time, last_ts, timestamp)
            self._packet["payload"]["stream_data"][i]["timestamp"] = reference_time * 1000
            last_timestamp[0] = reference_time
            last_timestamp[1] = timestamp


class EDADataPacket(CommandPacket):
    """
    Packet Structure:

    .. code-block::

        {
            'header': {
                'source': <Stream.EDA: ['0xC4', '0x02']>,
                'destination': <Application.APP_USB: ['0xC7', '0x05']>,
                'length': '0x3D',
                'checksum': '0x0'
            },
            'payload': {
                'command': <CommonCommand.STREAM_DATA: ['0x28']>,
                'status': <CommonStatus.OK: ['0x00']>,
                'sequence_number': 0,
                'data_type': 0,
                'stream_data': [
                    {
                        'timestamp': 1774366622,
                        'real_data': 0,
                        'imaginary_data': 0
                    },
                    {
                        'timestamp': 1774374407,
                        'real_data': 0,
                        'imaginary_data': 16708
                    },
                    {
                        'timestamp': 1774382157,
                        'real_data': 16728,
                        'imaginary_data': 3257
                    },
                    {
                        'timestamp': 1774389924,
                        'real_data': 3277,
                        'imaginary_data': -20751
                    },
                    {
                        'timestamp': 1774397691,
                        'real_data': -20731,
                        'imaginary_data': -15161
                    },
                    {
                        'timestamp': 1774405458,
                        'real_data': -15141,
                        'imaginary_data': -30319
                    }
                ]
            }
        }
    """

    def __init__(self):
        super().__init__()
        self._config["payload"]["sequence_number"] = {"size": 2, "join": True}
        self._config["payload"]["data_type"] = {"size": 1, "join": True}
        self._config["payload"]["timestamp1"] = {"size": 4, "join": True}
        self._config["payload"]["real_data1"] = {"size": 2, "join": True, "signed": True}
        self._config["payload"]["imaginary_data1"] = {"size": 2, "join": True, "signed": True}
        self._config["payload"]["timestamp2"] = {"size": 4, "join": True}
        self._config["payload"]["real_data2"] = {"size": 2, "join": True, "signed": True}
        self._config["payload"]["imaginary_data2"] = {"size": 2, "join": True, "signed": True}
        self._config["payload"]["timestamp3"] = {"size": 4, "join": True}
        self._config["payload"]["real_data3"] = {"size": 2, "join": True, "signed": True}
        self._config["payload"]["imaginary_data3"] = {"size": 2, "join": True, "signed": True}
        self._config["payload"]["timestamp4"] = {"size": 4, "join": True}
        self._config["payload"]["real_data4"] = {"size": 2, "join": True, "signed": True}
        self._config["payload"]["imaginary_data4"] = {"size": 2, "join": True, "signed": True}
        self._config["payload"]["timestamp5"] = {"size": 4, "join": True}
        self._config["payload"]["real_data5"] = {"size": 2, "join": True, "signed": True}
        self._config["payload"]["imaginary_data5"] = {"size": 2, "join": True, "signed": True}
        self._config["payload"]["timestamp6"] = {"size": 4, "join": True}
        self._config["payload"]["real_data6"] = {"size": 2, "join": True, "signed": True}
        self._config["payload"]["imaginary_data6"] = {"size": 2, "join": True, "signed": True}

    def decode_packet(self, data):
        super().decode_packet(data)
        self._packet["payload"]["stream_data"] = []
        for i in range(1, 7):
            timestamp = self._packet["payload"][f"timestamp{i}"]
            data = {"timestamp": timestamp,
                    "real_data": self._packet["payload"][f"real_data{i}"],
                    "imaginary_data": self._packet["payload"][f"imaginary_data{i}"]}
            [self._packet["payload"].pop(key) for key in [f"timestamp{i}", f"real_data{i}", f"imaginary_data{i}"]]
            self._packet["payload"]["stream_data"].append(data)

    def update_timestamp(self, last_timestamp):
        for i in range(len(self._packet["payload"]["stream_data"])):
            reference_time, last_ts = last_timestamp
            timestamp = self._packet["payload"]["stream_data"][i]["timestamp"]
            reference_time = utils.get_updated_timestamp(reference_time, last_ts, timestamp)
            self._packet["payload"]["stream_data"][i]["timestamp"] = reference_time * 1000
            last_timestamp[0] = reference_time
            last_timestamp[1] = timestamp


class PedometerDataPacket(CommandPacket):
    """
    Packet Structure:

    .. code-block::

        {
            'header': {
                'source': <Stream.PEDOMETER: ['0xC4', '0x04']>,
                'destination': <Application.APP_USB: ['0xC7', '0x05']>,
                'length': '0x17',
                'checksum': '0x0'
            },
            'payload': {
                'command': <CommonCommand.STREAM_DATA: ['0x28']>,
                'status': <CommonStatus.OK: ['0x00']>,
                'sequence_number': 1,
                'steps': 0,
                'algo_status': 8193,
                'timestamp': 1776836792,
                'reserved': 0
            }
        }
    """

    def __init__(self):
        super().__init__()
        self._config["payload"]["sequence_number"] = {"size": 2, "join": True}
        self._config["payload"]["steps"] = {"size": 4, "join": True, "signed": True}
        self._config["payload"]["algo_status"] = {"size": 2, "join": True}
        self._config["payload"]["timestamp"] = {"size": 4, "join": True}
        self._config["payload"]["reserved"] = {"size": 1, "join": True}

    def update_timestamp(self, last_timestamp):
        reference_time, last_ts = last_timestamp
        timestamp = self._packet["payload"]["timestamp"]
        reference_time = utils.get_updated_timestamp(reference_time, last_ts, timestamp)
        self._packet["payload"]["timestamp"] = reference_time * 1000
        last_timestamp[0] = reference_time
        last_timestamp[1] = timestamp


class TemperatureDataPacket(CommandPacket):
    """
    Packet Structure:

    .. code-block::

        {
            'header': {
                'source': <Stream.TEMPERATURE: ['0xC4', '0x06']>,
                'destination': <Application.APP_USB: ['0xC7', '0x05']>,
                'length': '0x14',
                'checksum': '0x0'
            },
            'payload': {
                'command': <CommonCommand.STREAM_DATA: ['0x28']>,
                'status': <CommonStatus.OK: ['0x00']>,
                'sequence_number': 2,
                'timestamp': 1779692742,
                'skin_temperature': 30.1, # celsius
                'impedance': 79000 # ohm
            }
        }
    """

    def __init__(self):
        super().__init__()
        self._config["payload"]["sequence_number"] = {"size": 2, "join": True}
        self._config["payload"]["timestamp"] = {"size": 4, "join": True}
        self._config["payload"]["skin_temperature"] = {"size": 2, "join": True}
        self._config["payload"]["impedance"] = {"size": 2, "join": True}

    def decode_packet(self, data):
        super().decode_packet(data)
        self._packet["payload"]["skin_temperature"] = self._packet["payload"]["skin_temperature"] / 10.0
        self._packet["payload"]["impedance"] = self._packet["payload"]["impedance"] * 100

    def update_timestamp(self, last_timestamp):
        reference_time, last_ts = last_timestamp
        timestamp = self._packet["payload"]["timestamp"]
        reference_time = utils.get_updated_timestamp(reference_time, last_ts, timestamp)
        self._packet["payload"]["timestamp"] = reference_time * 1000
        last_timestamp[0] = reference_time
        last_timestamp[1] = timestamp


class PPGDataPacket(CommandPacket):
    """
    Packet Structure:

    .. code-block::

        {
            'header': {
                'source': <Stream.PPG: ['0xC4', '0x00']>,
                'destination': <Application.APP_USB: ['0xC7', '0x05']>,
                'length': '0x2E',
                'checksum': '0x0'
            },
            'payload': {
                'command': <CommonCommand.STREAM_DATA: ['0x28']>,
                'status': <CommonStatus.NEW_STREAM_STATUS: ['0x43']>,
                'sequence_number': 9,
                'timestamp': 1782357951,
                'adpd_lib_state': 7,
                'hr': 0,
                'confidence': 0,
                'hr_type': 0,
                'rr_interval': 0,
                'debug_info': [ 0, 320, 50, 0, 5, 0, 320, 50, 978 ]
            }
        }
    """

    def __init__(self):
        super().__init__()
        self._config["payload"]["sequence_number"] = {"size": 2, "join": True}
        self._config["payload"]["timestamp"] = {"size": 4, "join": True}
        self._config["payload"]["adpd_lib_state"] = {"size": 2, "join": True}
        self._config["payload"]["hr"] = {"size": 2, "join": True}
        self._config["payload"]["confidence"] = {"size": 2, "join": True}
        self._config["payload"]["hr_type"] = {"size": 2, "join": True}
        self._config["payload"]["rr_interval"] = {"size": 2, "join": True}
        self._config["payload"]["debug_info"] = {"size": 2, "join": True}
        self._config["payload"]["debug_info1"] = {"size": 2, "join": True}
        self._config["payload"]["debug_info2"] = {"size": 2, "join": True}
        self._config["payload"]["debug_info3"] = {"size": 2, "join": True}
        self._config["payload"]["debug_info4"] = {"size": 2, "join": True}
        self._config["payload"]["debug_info5"] = {"size": 2, "join": True}
        self._config["payload"]["debug_info6"] = {"size": 2, "join": True}
        self._config["payload"]["debug_info7"] = {"size": 2, "join": True}
        self._config["payload"]["debug_info8"] = {"size": 2, "join": True}
        self._config["payload"]["debug_info9"] = {"size": 2, "join": True}

    def decode_packet(self, data):
        super().decode_packet(data)
        self._packet["payload"]["debug_info"] = [self._packet["payload"][f"debug_info"]]
        for i in range(1, 10):
            self._packet["payload"]["debug_info"].append(self._packet["payload"][f"debug_info{i}"])
            del self._packet["payload"][f"debug_info{i}"]

    def update_timestamp(self, last_timestamp):
        reference_time, last_ts = last_timestamp
        timestamp = self._packet["payload"]["timestamp"]
        reference_time = utils.get_updated_timestamp(reference_time, last_ts, timestamp)
        self._packet["payload"]["timestamp"] = reference_time * 1000
        last_timestamp[0] = reference_time
        last_timestamp[1] = timestamp


class SYNCPPGDataPacket(CommandPacket):
    """
    Packet Structure:

    .. code-block::

        {
            'header': {
                'source': <Stream.SYNC_PPG: ['0xC4', '0x05']>,
                'destination': <Application.APP_USB: ['0xC7', '0x05']>,
                'length': '0x48',
                'checksum': '0x0'
            },
            'payload': {
                'command': <CommonCommand.STREAM_DATA: ['0x28']>,
                'status': <CommonStatus.OK: ['0x00']>,
                'sequence_number': 119,
                'stream_data': [
                    {
                        'ppg_timestamp': 1782379042,
                        'ppg_data': 186278,
                        'adxl_timestamp': 1782378429,
                        'adxl_x': 48,
                        'adxl_y': 268,
                        'adxl_z': 62
                    },
                    {
                        'ppg_timestamp': 1782379682,
                        'ppg_data': 186278,
                        'adxl_timestamp': 1782379070,
                        'adxl_x': 55,
                        'adxl_y': 264,
                        'adxl_z': 60
                    },
                    {
                        'ppg_timestamp': 1782380323,
                        'ppg_data': 186289,
                        'adxl_timestamp': 1782379709,
                        'adxl_x': 68,
                        'adxl_y': 268,
                        'adxl_z': 57
                    },
                    {
                        'ppg_timestamp': 1782380962,
                        'ppg_data': 186305,
                        'adxl_timestamp': 1782380349,
                        'adxl_x': 60,
                        'adxl_y': 263,
                        'adxl_z': 62
                    }
                ]
            }
        }
    """

    def __init__(self):
        super().__init__()
        self._config["payload"]["sequence_number"] = {"size": 2, "join": True}
        self._config["payload"]["ppg_timestamp1"] = {"size": 4, "join": True}
        self._config["payload"]["adxl_timestamp1"] = {"size": 4, "join": True}
        self._config["payload"]["ppg_timestamp2"] = {"size": 2, "join": True}
        self._config["payload"]["ppg_timestamp3"] = {"size": 2, "join": True}
        self._config["payload"]["ppg_timestamp4"] = {"size": 2, "join": True}
        self._config["payload"]["adxl_timestamp2"] = {"size": 2, "join": True}
        self._config["payload"]["adxl_timestamp3"] = {"size": 2, "join": True}
        self._config["payload"]["adxl_timestamp4"] = {"size": 2, "join": True}
        self._config["payload"]["ppg_data1"] = {"size": 4, "join": True}
        self._config["payload"]["ppg_data2"] = {"size": 4, "join": True}
        self._config["payload"]["ppg_data3"] = {"size": 4, "join": True}
        self._config["payload"]["ppg_data4"] = {"size": 4, "join": True}
        self._config["payload"]["adxl_x1"] = {"size": 2, "join": True, "signed": True}
        self._config["payload"]["adxl_x2"] = {"size": 2, "join": True, "signed": True}
        self._config["payload"]["adxl_x3"] = {"size": 2, "join": True, "signed": True}
        self._config["payload"]["adxl_x4"] = {"size": 2, "join": True, "signed": True}
        self._config["payload"]["adxl_y1"] = {"size": 2, "join": True, "signed": True}
        self._config["payload"]["adxl_y2"] = {"size": 2, "join": True, "signed": True}
        self._config["payload"]["adxl_y3"] = {"size": 2, "join": True, "signed": True}
        self._config["payload"]["adxl_y4"] = {"size": 2, "join": True, "signed": True}
        self._config["payload"]["adxl_z1"] = {"size": 2, "join": True, "signed": True}
        self._config["payload"]["adxl_z2"] = {"size": 2, "join": True, "signed": True}
        self._config["payload"]["adxl_z3"] = {"size": 2, "join": True, "signed": True}
        self._config["payload"]["adxl_z4"] = {"size": 2, "join": True, "signed": True}

    def decode_packet(self, data):
        super().decode_packet(data)
        self._packet["payload"]["stream_data"] = []
        ppg_timestamp = 0
        adxl_timestamp = 0
        for i in range(1, 5):
            ppg_timestamp += self._packet["payload"][f"ppg_timestamp{i}"]
            adxl_timestamp += self._packet["payload"][f"adxl_timestamp{i}"]
            data = {"ppg_timestamp": ppg_timestamp,
                    "ppg_data": self._packet["payload"][f"ppg_data{i}"],
                    "adxl_timestamp": adxl_timestamp,
                    "adxl_x": self._packet["payload"][f"adxl_x{i}"],
                    "adxl_y": self._packet["payload"][f"adxl_y{i}"],
                    "adxl_z": self._packet["payload"][f"adxl_z{i}"]}
            [self._packet["payload"].pop(key) for key in
             [f"ppg_timestamp{i}", f"ppg_data{i}", f"adxl_timestamp{i}", f"adxl_x{i}", f"adxl_y{i}", f"adxl_z{i}"]]
            self._packet["payload"]["stream_data"].append(data)

    def update_timestamp(self, last_timestamp):
        for i in range(len(self._packet["payload"]["stream_data"])):
            ppg_reference_time, ppg_last_ts, adxl_reference_time, adxl_last_ts = last_timestamp
            ppg_timestamp = self._packet["payload"]["stream_data"][i]["ppg_timestamp"]
            adxl_timestamp = self._packet["payload"]["stream_data"][i]["adxl_timestamp"]
            ppg_reference_time = utils.get_updated_timestamp(ppg_reference_time, ppg_last_ts, ppg_timestamp)
            adxl_reference_time = utils.get_updated_timestamp(adxl_reference_time, adxl_last_ts, adxl_timestamp)
            self._packet["payload"]["stream_data"][i]["ppg_timestamp"] = ppg_reference_time * 1000
            self._packet["payload"]["stream_data"][i]["adxl_timestamp"] = adxl_reference_time * 1000
            last_timestamp[0] = ppg_reference_time
            last_timestamp[1] = ppg_timestamp
            last_timestamp[2] = adxl_reference_time
            last_timestamp[3] = adxl_timestamp


class SQIDataPacket(CommandPacket):
    """
    Packet Structure:

    .. code-block::

        {
            'header': {
                'source': <Stream.SQI: ['0xC8', '0x0D']>,
                'destination': <Application.APP_USB: ['0xC7', '0x05']>,
                'length': '0x19',
                'checksum': '0x0'
            },
            'payload': {
                'command': <CommonCommand.STREAM_DATA: ['0x28']>,
                'status': <CommonStatus.OK: ['0x00']>,
                'sequence_number': 0,
                'sqi': 2.537532566293521e-07,
                'sqi_slot': 42405,
                'algo_status': 0,
                'timestamp': 1786433199,
                'reserved': 0
            }
        }
    """

    def __init__(self):
        super().__init__()
        self._config["payload"]["sequence_number"] = {"size": 2, "join": True}
        self._config["payload"]["sqi"] = {"size": 4}
        self._config["payload"]["sqi_slot"] = {"size": 2, "join": True}
        self._config["payload"]["algo_status"] = {"size": 2, "join": True}
        self._config["payload"]["timestamp"] = {"size": 4, "join": True}
        self._config["payload"]["reserved"] = {"size": 1, "join": True}

    def decode_packet(self, data):
        super().decode_packet(data)
        self._packet["payload"]["sqi"] = struct.unpack("f", bytes(self._packet["payload"]["sqi"]))[0]

    def update_timestamp(self, last_timestamp):
        reference_time, last_ts = last_timestamp
        timestamp = self._packet["payload"]["timestamp"]
        reference_time = utils.get_updated_timestamp(reference_time, last_ts, timestamp)
        self._packet["payload"]["timestamp"] = reference_time * 1000
        last_timestamp[0] = reference_time
        last_timestamp[1] = timestamp


class BCMDataPacket(CommandPacket):
    """
    Packet Structure:

    .. code-block::

        {
            'header': {
                'source': <Stream.BCM: ['0xC4', '0x07']>,
                'destination': <Application.APP_USB: ['0xC7', '0x05']>,
                'length': '0x41',
                'checksum': '0x0'
            },
            'payload': {
                'command': <CommonCommand.STREAM_DATA: ['0x28']>,
                'status': <CommonStatus.OK: ['0x00']>,
                'sequence_number': 1,
                'data_type': 0,
                'stream_data': [
                    {
                        'timestamp': 1788107093,
                        'real': 0,
                        'imaginary': 0,
                        'frequency_index': 82
                    },
                    {
                        'timestamp': 1788114863,
                        'real': 0,
                        'imaginary': 0,
                        'frequency_index': 5
                    },
                    {
                        'timestamp': 1788122630,
                        'real': 0,
                        'imaginary': 0,
                        'frequency_index': 8
                    },
                    {
                        'timestamp': 1788130399,
                        'real': 0,
                        'imaginary': 0,
                        'frequency_index': 54
                    }
                ]
            }
        }
    """

    def __init__(self):
        super().__init__()
        self._config["payload"]["sequence_number"] = {"size": 2, "join": True}
        self._config["payload"]["data_type"] = {"size": 1, "join": True}
        self._config["payload"]["timestamp1"] = {"size": 4, "join": True}
        self._config["payload"]["real1"] = {"size": 4, "join": True, "signed": True}
        self._config["payload"]["imaginary1"] = {"size": 4, "join": True, "signed": True}
        self._config["payload"]["frequency_index1"] = {"size": 1, "join": True}
        self._config["payload"]["timestamp2"] = {"size": 4, "join": True}
        self._config["payload"]["real2"] = {"size": 4, "join": True, "signed": True}
        self._config["payload"]["imaginary2"] = {"size": 4, "join": True, "signed": True}
        self._config["payload"]["frequency_index2"] = {"size": 1, "join": True}
        self._config["payload"]["timestamp3"] = {"size": 4, "join": True}
        self._config["payload"]["real3"] = {"size": 4, "join": True, "signed": True}
        self._config["payload"]["imaginary3"] = {"size": 4, "join": True, "signed": True}
        self._config["payload"]["frequency_index3"] = {"size": 1, "join": True}
        self._config["payload"]["timestamp4"] = {"size": 4, "join": True}
        self._config["payload"]["real4"] = {"size": 4, "join": True, "signed": True}
        self._config["payload"]["imaginary4"] = {"size": 4, "join": True, "signed": True}
        self._config["payload"]["frequency_index4"] = {"size": 1, "join": True}

    def decode_packet(self, data):
        super().decode_packet(data)
        self._packet["payload"]["stream_data"] = []
        for i in range(1, 5):
            timestamp = self._packet["payload"][f"timestamp{i}"]
            data = {"timestamp": timestamp,
                    "real": self._packet["payload"][f"real{i}"],
                    "imaginary": self._packet["payload"][f"imaginary{i}"],
                    "frequency_index": self._packet["payload"][f"frequency_index{i}"]}
            [self._packet["payload"].pop(key) for key in
             [f"timestamp{i}", f"real{i}", f"imaginary{i}", f"frequency_index{i}"]]
            self._packet["payload"]["stream_data"].append(data)

    def update_timestamp(self, last_timestamp):
        for i in range(len(self._packet["payload"]["stream_data"])):
            reference_time, last_ts = last_timestamp
            timestamp = self._packet["payload"]["stream_data"][i]["timestamp"]
            reference_time = utils.get_updated_timestamp(reference_time, last_ts, timestamp)
            self._packet["payload"]["stream_data"][i]["timestamp"] = reference_time * 1000
            last_timestamp[0] = reference_time
            last_timestamp[1] = timestamp


class KeyStreamDataPacket(CommandPacket):
    """
    Packet Structure:

    .. code-block::

        {
            'header': {
                'source': <Application.DISPLAY: ['0xC5', '0x03']>,
                'destination': <Application.APP_USB: ['0xC7', '0x05']>,
                'length': '0xB',
                'checksum': '0x0'
            },
            'payload': {
                'command': <DisplayCommand.KEY_STREAM_DATA: ['0x48']>,
                'status': <CommonStatus.OK: ['0x00']>,
                'key_code': 18
            }
        }
    """

    def __init__(self, destination=None, command=None):
        super().__init__(destination, command)
        self._config["payload"]["key_code"] = {"size": 1, "join": True}

    def update_timestamp(self, last_timestamp):
        pass


class CapSenseStreamDataPacket(CommandPacket):
    """
    Packet Structure:

    .. code-block::

        {
            'header': {
                'source': <Application.PM: ['0xC5', '0x00']>,
                'destination': <Application.APP_USB: ['0xC7', '0x05']>,
                'length': '0xC',
                'checksum': '0x0'
            },
            'payload': {
                'command': <PMCommand.CAP_SENSE_STREAM_DATA: ['0x82']>,
                'status': <CommonStatus.OK: ['0x00']>,
                'position': 1,
                'value': 0
            }
        }
    """

    def __init__(self, destination=None, command=None):
        super().__init__(destination, command)
        self._config["payload"]["position"] = {"size": 1, "join": True}
        self._config["payload"]["value"] = {"size": 1, "join": True}

    def update_timestamp(self, last_timestamp):
        pass



# ******************************************************************************
# Copyright (c) 2019 Analog Devices, Inc.  All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# - Redistributions of source code must retain the above copyright notice, this
#  list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice,
#  this list of conditions and the following disclaimer in the documentation
#  and/or other materials provided with the distribution.
# - Modified versions of the software must be conspicuously marked as such.
# - This software is licensed solely and exclusively for use with
#  processors/products manufactured by or for Analog Devices, Inc.
# - This software may not be combined or merged with other code in any manner
#  that would cause the software to become subject to terms and conditions
#  which differ from those listed here.
# - Neither the name of Analog Devices, Inc. nor the names of its contributors
#  may be used to endorse or promote products derived from this software
#  without specific prior written permission.
# - The use of this software may or may not infringe the patent rights of one
#  or more patent holders.  This license does not release you from the
#  requirement that you obtain separate licenses from these patent holders to
#  use this software.
#
# THIS SOFTWARE IS PROVIDED BY ANALOG DEVICES, INC. AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# NONINFRINGEMENT, TITLE, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL ANALOG DEVICES, INC. OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, PUNITIVE OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, DAMAGES ARISING OUT OF
# CLAIMS OF INTELLECTUAL PROPERTY RIGHTS INFRINGEMENT; PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# ******************************************************************************

import os
import logging
import threading
import time
from queue import Queue

from tqdm import tqdm

from . import utils
from .enums.common_enums import Application
from .enums.low_touch_enum import CommandType

logger = logging.getLogger(__name__)


class PacketManager:
    """

    PacketManager

    """

    def __init__(self, serial_object, filename=None):
        """
        Initialization of packet Manager

        :param serial_object: Serial port object
        :type serial_object: Serial
        """
        self.count = 0
        self.queue = Queue()
        self.filename = filename
        self.command_type = None
        self.subscribe_source = {}
        self.processing_file = True
        self.lock = threading.Lock()
        self.command_logging = None
        self.source = Application.APP_USB
        self.serial_object = serial_object
        self._kill_threads = threading.Event()
        self.receive_thread = threading.Thread(target=self.receive_packet)
        self.file_thread = threading.Thread(target=self.process_file)
        self.log_command_packet = None

    def set_usb_source(self):
        self.source = Application.APP_USB

    def set_ble_source(self):
        self.source = Application.APP_BLE

    def enable_command_logging(self, command_type):
        self.command_logging = True
        self.command_type = command_type

    def disable_command_logging(self, command_type):
        self.command_logging = False
        self.command_type = command_type

    def subscribe_command_logger(self, log_command):
        self.log_command_packet = log_command

    def unsubscribe_command_logger(self):
        self.command_type = None
        self.command_logging = None
        self.log_command_packet = None

    def process_queue(self):
        """
        Reads the packet from queue and send to relevant callback function.
        """
        while True:
            try:
                pkt = self.queue.get()
                if self._kill_threads.isSet():
                    break
                pkt = list(pkt)
                logger.info(f"RX::{':'.join(utils.convert_int_array_to_hex(pkt))}")
                packet_id = pkt[0] + (pkt[1] << 8) + (pkt[8] << 16)
                callbacks = self.subscribe_source.get(packet_id, [])
                if len(callbacks):
                    for callback in callbacks:
                        callback(pkt, packet_id)
                else:
                    logger.warning(f"Received an unexpected packet {pkt} with packet ID {packet_id}")
            except Exception as e:
                logger.critical(f"Process_queue thread failed : {e}.", exc_info=True)
                break

    def receive_packet(self):
        """
        Reads packet from serial port and store it in queue.
        """
        while True:
            try:
                if self._kill_threads.isSet():
                    break
                pkt = self.serial_object.read(8)
                if len(pkt):
                    length = (pkt[4] << 8) + pkt[5]
                    pkt += self.serial_object.read(length - 8)
                    self.queue.put(pkt)
            except Exception as e:
                logger.critical(f"Receive_packet thread failed : {e}.", exc_info=True)
                break

    def process_file(self):
        """
        Reads packet from file and store it in file queue.
        """
        if self.filename:
            file_size = os.path.getsize(self.filename)
            progress_bar = tqdm(total=file_size)
            with open(self.filename, 'rb') as file:
                i = 1
                while True:
                    if self._kill_threads.isSet():
                        break
                    try:
                        pkt = file.read(8)
                        if len(pkt):
                            length = (int(pkt[4]) << 8) + int(pkt[5])
                            progress_bar.update(length)
                            pkt += file.read(length - 8)
                            self.queue.put(pkt)
                            if i == 1:
                                time.sleep(1)
                                i = 2
                        else:
                            break
                    except Exception as e:
                        logger.critical(f"Receive_packet thread failed : {e}.", exc_info=True)
                        break
                progress_bar.close()
                self.processing_file = False

    def send_packet(self, packet):
        """
        Receives packet from sdk and send it to the serial to port for device to process.
        """
        packet.set_source(self.source)
        packet.update_packet_length()
        logger.debug(f"Packet ready : {packet}")
        packet = packet.to_list()
        # if not convert_log_to_csv
        if not self.filename:
            self.serial_object.write(packet)
        # logging for LT enable_command_logging
        if self.command_logging is True and self.command_type == CommandType.START:
            self.log_command_packet.add_start_command(packet)
        elif self.command_logging is True and self.command_type == CommandType.STOP:
            self.log_command_packet.add_stop_command(packet)
        logger.info(f"TX :: {':'.join(utils.convert_int_array_to_hex(packet))}")

    def start_receive_and_process_threads(self):
        """
        start process and receive thread.
        """
        if self.filename:
            self.file_thread.setDaemon(True)
            self.file_thread.start()
        else:
            self.receive_thread.setDaemon(True)
            self.receive_thread.start()
        thread = threading.Thread(target=self.process_queue)
        thread.setDaemon(True)
        thread.name = f"process_thread"
        thread.start()

    def subscribe(self, packet_id, callback):
        """
        Subscribe packet id to the callback.
        """
        callbacks = self.subscribe_source.get(packet_id, [])
        callbacks.append(callback)
        self.subscribe_source[packet_id] = callbacks

    def unsubscribe(self, packet_id, application_callback):
        """
        Unsubscribe packet id and removes the callback.
        """
        callbacks = self.subscribe_source.get(packet_id, [])
        for i, callback in enumerate(callbacks):
            if callback == application_callback:
                del callbacks[i]
                break
        self.subscribe_source[packet_id] = callbacks

    def close(self):
        """Closes serial port"""
        self._kill_threads.set()
        # to free up queue.get() lock in process_queue add -1 to queue
        self.queue.put(-1)
        self.serial_object.close()



# ******************************************************************************
# Copyright (c) 2019 Analog Devices, Inc.  All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# - Redistributions of source code must retain the above copyright notice, this
#  list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice,
#  this list of conditions and the following disclaimer in the documentation
#  and/or other materials provided with the distribution.
# - Modified versions of the software must be conspicuously marked as such.
# - This software is licensed solely and exclusively for use with
#  processors/products manufactured by or for Analog Devices, Inc.
# - This software may not be combined or merged with other code in any manner
#  that would cause the software to become subject to terms and conditions
#  which differ from those listed here.
# - Neither the name of Analog Devices, Inc. nor the names of its contributors
#  may be used to endorse or promote products derived from this software
#  without specific prior written permission.
# - The use of this software may or may not infringe the patent rights of one
#  or more patent holders.  This license does not release you from the
#  requirement that you obtain separate licenses from these patent holders to
#  use this software.
#
# THIS SOFTWARE IS PROVIDED BY ANALOG DEVICES, INC. AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# NONINFRINGEMENT, TITLE, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL ANALOG DEVICES, INC. OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, PUNITIVE OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, DAMAGES ARISING OUT OF
# CLAIMS OF INTELLECTUAL PROPERTY RIGHTS INFRINGEMENT; PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# ******************************************************************************

import logging

logger = logging.getLogger(__name__)


def get_updated_timestamp(reference_time, last_ts, timestamp):
    if last_ts > timestamp:
        change = (timestamp + 2764800000.0) - last_ts
    else:
        change = timestamp - last_ts
    change = change / 32000.0
    return reference_time + change


def join_multi_length_packets(packet, sign=False, reverse=False, convert_to_hex=False):
    """
    Joins array of bytes into integer.
    """
    ans = 0
    packet_len = len(packet)
    if packet_len == 0:
        return ans
    if reverse:
        packet = list(reversed(packet))
    for i, value in enumerate(packet):
        ans += (value << (8 * i))
    bits = packet_len * 8
    if sign and ans & (1 << (bits - 1)):
        ans -= 1 << bits
    if convert_to_hex:
        return "0x%X" % ans
    return ans


def range_and_type_check(num, num_bytes=None, lower_and_upper_bound=None, type_of=None, max_len=None):
    """
    Checks num for specified type and range.
    """
    if type_of:
        if not type(num) == type_of:
            logger.error(f"{'0x%X' % num} is not of type {type_of}.")

    if num_bytes:
        lower_bound = 0
        upper_bound = 16 ** (num_bytes * 2) - 1
        if not lower_bound <= num <= upper_bound:
            logger.error(f"{'0x%X' % num} is out of range, Variable needs to be between {'0x%X' % lower_bound}"
                         f" and {'0x%X' % upper_bound}.")

    if lower_and_upper_bound:
        if not lower_and_upper_bound[0] <= num <= lower_and_upper_bound[1]:
            logger.error(f"{'0x%X' % num} is out of range, Variable needs to be greater than or equal to "
                         f"{'0x%X' % lower_and_upper_bound[0]} and less than or equal to "
                         f"{'0x%X' % lower_and_upper_bound[1]}.")

    if max_len:
        if len(num) > max_len:
            logger.error(f"{num} is out of range. Max length allowed is {max_len}.")
    return num


def split_int_in_bytes(value, length=None, reverse=False):
    """
    Breaks int into array of byte array of specified length.
    """
    result = []
    shift = 0
    base_value = 0
    if value < 0:
        base_value = 255
    while value >> shift and not (value < 0 and value >> shift == -1):
        result.append((value >> shift) & 0xff)
        shift += 8
    if length and not len(result) == length:
        result += [base_value] * abs(length - len(result))
    if reverse:
        result = list(reversed(result))
    return result


def address_range_check(address, address_range):
    """
    checks if address is in given range or not.
    """
    if not address_range[0] <= address <= address_range[1]:
        logger.warning(
            f"Address {'0x%X' % address} is out of range, it should be in range {'0x%X' % address_range[0]} "
            f"and {'0x%X' % address_range[1]} ")
    return address


def check_array_address_range(array, address_range, num_bytes, type_of=int):
    """
    checks if array address is in given range or not.
    """
    for values in array:
        if type(values) == list:
            address_range_check(values[0], address_range)
            range_and_type_check(values[1], type_of=type_of, num_bytes=num_bytes)
        else:
            address_range_check(values, address_range)
            range_and_type_check(values, type_of=type_of, num_bytes=num_bytes)


def convert_int_array_to_hex(arr):
    """
    Convert int to hex.
    """
    return ['0x%02X' % x for x in arr]


def pretty(value, tab_char='\t', next_line_char='\n', indent=0):
    """
    Print dict in clean format.
    """
    line = next_line_char + tab_char * (indent + 1)
    if type(value) is dict:
        items = [
            line + repr(key) + ': ' + pretty(value[key], tab_char, next_line_char, indent + 1)
            for key in value
        ]
        return '{%s}' % (','.join(items) + next_line_char + tab_char * indent)
    elif type(value) is list:
        items = []
        flag = " "
        for item in value:
            if type(item) is list:
                flag = line[:-1]
                items.append(line + pretty(item, tab_char, next_line_char, indent))
            elif type(item) is dict:
                flag = line[:-1]
                items.append(line + pretty(item, tab_char, next_line_char, indent + 1))
            else:
                items.append(" " + pretty(item, tab_char, " ", 0))
        return '[%s%s]' % (','.join(items), flag)
    elif type(value) is tuple:
        items = [
            line + pretty(item, tab_char, next_line_char, indent + 1)
            for item in value
        ]
        return '(%s)' % (','.join(items) + next_line_char + tab_char * indent)
    else:
        return repr(value)
